__author__ = 'maartenbreddels'
import os
import sys
import collections
import numpy as np
import logging
import concurrent.futures
import threading

import vaex
import vaex.utils
import vaex.execution
import vaex.file.colfits
import vaex.file.other
from vaex.column import ColumnStringArrow, str_type, _to_string_sequence


max_length = int(1e5)

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
try:
    import h5py
except:
    if not on_rtd:
        raise
# from vaex.dataset import DatasetLocal

logger = logging.getLogger("vaex.export")
progress_lock = threading.Lock()

class ProgressStatus(object):
    pass

def _export(dataset_input, dataset_output, random_index_column, path, column_names=None, byteorder="=", shuffle=False, selection=False, progress=None, virtual=True, sort=None, ascending=True):
    """
    :param DatasetLocal dataset: dataset to export
    :param str path: path for file
    :param lis[str] column_names: list of column names to export or None for all columns
    :param str byteorder: = for native, < for little endian and > for big endian
    :param bool shuffle: export rows in random order
    :param bool selection: export selection or not
    :param progress: progress callback that gets a progress fraction as argument and should return True to continue
    :return:
    """

    if selection:
        if selection == True:  # easier to work with the name
            selection = "default"

    N = len(dataset_input) if not selection else dataset_input.selected_length(selection)
    if N == 0:
        raise ValueError("Cannot export empty table")

    if shuffle and sort:
        raise ValueError("Cannot shuffle and sort at the same time")

    if shuffle:
        shuffle_array = dataset_output.columns[random_index_column]

    partial_shuffle = shuffle and len(dataset_input) != N

    order_array = None
    order_array_inverse = None

    # for strings we also need the inverse order_array, keep track of that
    has_strings = any([dataset_input.dtype(k) == str_type for k in column_names])

    if partial_shuffle:
        # if we only export a portion, we need to create the full length random_index array, and
        shuffle_array_full = np.random.choice(len(dataset_input), len(dataset_input), replace=False)
        # then take a section of it
        shuffle_array[:] = shuffle_array_full[shuffle_array_full < N]
        del shuffle_array_full
        order_array = shuffle_array
    elif shuffle:
        # better to do this in memory
        shuffle_array_memory = np.random.choice(N, N, replace=False)
        shuffle_array[:] = shuffle_array_memory
        order_array = shuffle_array
    if order_array is not None:
        indices_r = np.zeros_like(order_array)
        indices_r[order_array] = np.arange(len(order_array))
        order_array_inverse = indices_r
        del indices_r

    if sort:
        if selection:
            raise ValueError("sorting selections not yet supported")
        # these indices sort the input array, but we evaluate the input in sequential order and write it out in sorted order
        # e.g., not b[:] = a[indices]
        # but b[indices_r] = a
        logger.info("sorting...")
        indices = np.argsort(dataset_input.evaluate(sort))
        indices_r = np.zeros_like(indices)
        indices_r[indices] = np.arange(len(indices))
        if has_strings:
            # in this case we already have the inverse ready
            order_array_inverse = indices if ascending else indices[:--1]
        else:
            del indices
        order_array = indices_r if ascending else indices_r[::-1]
        logger.info("sorting done")

    if progress == True:
        progress = vaex.utils.progressbar_callable(title="exporting")
    progress = progress or (lambda value: True)
    progress_total = len(column_names) * len(dataset_input)
    progress_status = ProgressStatus()
    progress_status.cancelled = False
    progress_status.value = 0
    if selection:
        full_mask = dataset_input.evaluate_selection_mask(selection)
    else:
        full_mask = None

    sparse_groups = collections.defaultdict(list)
    sparse_matrices = {}  # alternative to a set of matrices, since they are not hashable
    string_columns = []
    futures = []

    thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=1) 
    if True:
        for column_name in column_names:
            sparse_matrix = dataset_output._sparse_matrix(column_name)
            if sparse_matrix is not None:
                # sparse columns are written differently
                sparse_groups[id(sparse_matrix)].append(column_name)
                sparse_matrices[id(sparse_matrix)] = sparse_matrix
                continue
            logger.debug("  exporting column: %s " % column_name)
            future = thread_pool.submit(_export_column, dataset_input, dataset_output, column_name, full_mask,
                shuffle, sort, selection, N, order_array, order_array_inverse, progress_status)
            futures.append(future)

    done = False
    while not done:
        done = True
        for future in futures:
            try:
                future.result(0.1/4)
            except concurrent.futures.TimeoutError:
                done = False
                break
        if not done:
            if not progress(progress_status.value / float(progress_total)):
                progress_status.cancelled = True

    for sparse_matrix_id, column_names in sparse_groups.items():
        sparse_matrix = sparse_matrices[sparse_matrix_id]
        for column_name in column_names:
            assert not shuffle
            assert selection in [None, False]
            column = dataset_output.columns[column_name]
            column.matrix.data[:] = dataset_input.columns[column_name].matrix.data
            column.matrix.indptr[:] = dataset_input.columns[column_name].matrix.indptr
            column.matrix.indices[:] = dataset_input.columns[column_name].matrix.indices
    return column_names

def _export_column(dataset_input, dataset_output, column_name, full_mask, shuffle, sort, selection, N, 
    order_array, order_array_inverse, progress_status):

        if 1:
            block_scope = dataset_input._block_scope(0, vaex.execution.buffer_size_default)
            to_array = dataset_output.columns[column_name]
            dtype = dataset_input.dtype(column_name)
            if shuffle or sort:  # we need to create a in memory copy, otherwise we will do random writes which is VERY inefficient
                to_array_disk = to_array
                if np.ma.isMaskedArray(to_array):
                    to_array = np.empty_like(to_array_disk)
                else:
                    if dtype == str_type:
                        # we create an empty column copy
                        to_array = to_array._zeros_like()
                    else:
                        to_array = np.zeros_like(to_array_disk)
            to_offset = 0  # we need this for selections
            count = len(dataset_input) if not selection else dataset_input.length_unfiltered()
            is_string = dtype == str_type
            # TODO: if no filter, selection or mask, we can choose the quick path for str
            string_byte_offset = 0

            for i1, i2 in vaex.utils.subdivide(count, max_length=max_length):
                logger.debug("from %d to %d (total length: %d, output length: %d)", i1, i2, len(dataset_input), N)
                block_scope.move(i1, i2)
                if selection:
                    mask = full_mask[i1:i2]
                    values = dataset_input.evaluate(column_name, i1=i1, i2=i2, filtered=False) #selection=selection)
                    values = values[mask]
                    no_values = len(values)
                    if no_values:
                        if is_string:
                            to_column = to_array
                            assert isinstance(to_column, ColumnStringArrow)
                            from_sequence = _to_string_sequence(values)
                            to_sequence = to_column.string_sequence.slice(to_offset, to_offset+no_values, string_byte_offset)
                            string_byte_offset += to_sequence.fill_from(from_sequence)
                            to_offset += no_values
                        else:
                            fill_value = np.nan if dtype.kind == "f" else None
                            # assert np.ma.isMaskedArray(to_array) == np.ma.isMaskedArray(values), "to (%s) and from (%s) array are not of both masked or unmasked (%s)" %\
                            # (np.ma.isMaskedArray(to_array), np.ma.isMaskedArray(values), column_name)
                            if dtype.type == np.datetime64:
                                values = values.view(np.int64)
                            if np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
                                to_array.data[to_offset:to_offset + no_values] = values.filled(fill_value)
                                to_array.mask[to_offset:to_offset + no_values] = values.mask
                            elif not np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
                                to_array[to_offset:to_offset + no_values] = values.filled(fill_value)
                            else:
                                to_array[to_offset:to_offset + no_values] = values
                            to_offset += no_values
                else:
                    values = dataset_input.evaluate(column_name, i1=i1, i2=i2)
                    if is_string:
                        no_values = len(values)
                        # for strings, we don't take sorting/shuffling into account when building the structure
                        to_column = to_array
                        assert isinstance(to_column, ColumnStringArrow)
                        from_sequence = _to_string_sequence(values)
                        to_sequence = to_column.string_sequence.slice(i1, i2, string_byte_offset)
                        string_byte_offset += to_sequence.fill_from(from_sequence)
                    else:
                        assert len(values) == (i2-i1)
                        fill_value = np.nan if dtype.kind == "f" else None
                        # assert np.ma.isMaskedArray(to_array) == np.ma.isMaskedArray(values), "to (%s) and from (%s) array are not of both masked or unmasked (%s)" %\
                        # (np.ma.isMaskedArray(to_array), np.ma.isMaskedArray(values), column_name)
                        if dtype.type == np.datetime64:
                            values = values.view(np.int64)
                        if shuffle or sort:
                            indices = order_array[i1:i2]
                            if np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
                                to_array.data[indices] = values.filled(fill_value)
                                to_array.mask[indices] = values.mask
                            elif not np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
                                to_array[indices] = values.filled(fill_value)
                            else:
                                to_array[indices] = values
                        else:
                            if np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
                                to_array.data[i1:i2] = values.filled(fill_value)
                                to_array.mask[i1:i2] = values.mask
                            elif np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
                                to_array[i1:i2] = values.filled(fill_value)
                            else:
                                to_array[i1:i2] = values
                with progress_lock:
                    progress_status.value += i2 - i1
                if progress_status.cancelled:
                    break
                #if not progress(progress_value / float(progress_total)):
                #    break
            if is_string:  # write out the last index
                to_column = to_array
                if selection:
                    to_column.indices[to_offset] = string_byte_offset
                else:
                    to_column.indices[count] = string_byte_offset
            if shuffle or sort:  # write to disk in one go
                if dtype == str_type:  # strings are sorted afterwards
                    view = to_array.string_sequence.lazy_index(order_array_inverse)
                    to_array_disk.string_sequence.fill_from(view)
                else:
                    if np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(to_array_disk):
                        to_array_disk.data[:] = to_array.data
                        to_array_disk.mask[:] = to_array.mask
                    else:
                        to_array_disk[:] = to_array


def export_fits(dataset, path, column_names=None, shuffle=False, selection=False, progress=None, virtual=True, sort=None, ascending=True):
    """
    :param DatasetLocal dataset: dataset to export
    :param str path: path for file
    :param lis[str] column_names: list of column names to export or None for all columns
    :param bool shuffle: export rows in random order
    :param bool selection: export selection or not
    :param progress: progress callback that gets a progress fraction as argument and should return True to continue,
            or a default progress bar when progress=True
    :param: bool virtual: When True, export virtual columns
    :return:
    """
    if shuffle:
        random_index_name = "random_index"
        while random_index_name in dataset.get_column_names():
            random_index_name += "_new"

    column_names = column_names or dataset.get_column_names(virtual=virtual, strings=True)
    logger.debug("exporting columns(fits): %r" % column_names)
    N = len(dataset) if not selection else dataset.selected_length(selection)
    data_types = []
    data_shapes = []
    ucds = []
    units = []
    for column_name in column_names:
        if column_name in dataset.get_column_names(strings=True):
            column = dataset.columns[column_name]
            shape = (N,) + column.shape[1:]
            dtype = column.dtype
            if dataset.dtype(column_name) == str_type:
                max_length = dataset[column_name].apply(lambda x: len(x)).max(selection=selection)
                dtype = np.dtype('S'+str(int(max_length)))
        else:
            dtype = np.float64().dtype
            shape = (N,)
        ucds.append(dataset.ucds.get(column_name))
        units.append(dataset.units.get(column_name))
        data_types.append(dtype)
        data_shapes.append(shape)

    if shuffle:
        column_names.append(random_index_name)
        data_types.append(np.int64().dtype)
        data_shapes.append((N,))
        ucds.append(None)
        units.append(None)
    else:
        random_index_name = None

    # TODO: all expressions can have missing values.. how to support that?
    null_values = {key: dataset.columns[key].fill_value for key in dataset.get_column_names() if dataset.is_masked(key) and dataset.dtype(key).kind != "f"}
    vaex.file.colfits.empty(path, N, column_names, data_types, data_shapes, ucds, units, null_values=null_values)
    if shuffle:
        del column_names[-1]
        del data_types[-1]
        del data_shapes[-1]
    dataset_output = vaex.file.other.FitsBinTable(path, write=True)
    _export(dataset_input=dataset, dataset_output=dataset_output, path=path, random_index_column=random_index_name,
            column_names=column_names, selection=selection, shuffle=shuffle,
            progress=progress, sort=sort, ascending=ascending)
    dataset_output.close_files()


def export_hdf5_v1(dataset, path, column_names=None, byteorder="=", shuffle=False, selection=False, progress=None, virtual=True):
    kwargs = locals()
    import vaex.hdf5.export
    vaex.hdf5.export.export_hdf5_v1(**kwargs)


def export_hdf5(dataset, path, column_names=None, byteorder="=", shuffle=False, selection=False, progress=None, virtual=True, sort=None, ascending=True):
    kwargs = locals()
    import vaex.hdf5.export
    vaex.hdf5.export.export_hdf5(**kwargs)


if __name__ == "__main__":
    sys.exit(main(sys.argv))


def main(argv):
    import argparse
    parser = argparse.ArgumentParser(argv[0])
    parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument('--quiet', '-q', default=False, action='store_true', help="do not output anything")
    parser.add_argument('--list', '-l', default=False, action='store_true', help="list columns of input")
    parser.add_argument('--progress', help="show progress (default: %(default)s)", default=True, action='store_true')
    parser.add_argument('--no-progress', dest="progress", action='store_false')
    parser.add_argument('--shuffle', "-s", dest="shuffle", action='store_true', default=False)
    parser.add_argument('--sort', dest="sort", default=None)
    parser.add_argument('--virtual', dest="virtual", action='store_true', default=False, help="Also export virtual columns")
    parser.add_argument('--fraction', "-f", dest="fraction", type=float, default=1.0, help="fraction of input dataset to export")
    parser.add_argument('--filter', dest="filter", default=None, help="filter to apply before exporting")

    subparsers = parser.add_subparsers(help='type of input source', dest="task")

    parser_soneira = subparsers.add_parser('soneira', help='create soneira peebles dataset')
    parser_soneira.add_argument('output', help='output file')
    parser_soneira.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")
    parser_soneira.add_argument('--dimension', '-d', type=int, help='dimensions', default=4)
    # parser_soneira.add_argument('--eta','-e', type=int, help='dimensions', default=3)
    parser_soneira.add_argument('--max-level', '-m', type=int, help='dimensions', default=28)
    parser_soneira.add_argument('--lambdas', '-l', type=int, help='lambda values for fractal', default=[1.1, 1.3, 1.6, 2.])

    parser_tap = subparsers.add_parser('tap', help='use TAP (Table Access Protocol) as source')
    parser_tap.add_argument("tap_url", help="input source or file")
    parser_tap.add_argument("table_name", help="input source or file")
    parser_tap.add_argument("output", help="output file (ends in .fits or .hdf5)")
    parser_tap.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")

    parser_file = subparsers.add_parser('file', help='use a file as source (e.g. .hdf5, .fits, .vot (VO table), .asc (ascii)')
    parser_file.add_argument("input", help="input source or file, when prefixed with @ it is assumed to be a text file with a file list (one file per line)")
    parser_file.add_argument("output", help="output file (ends in .fits or .hdf5)")
    parser_file.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")

    parser_file = subparsers.add_parser('csv', help='use a csv file as source (e.g. .hdf5, .fits, .vot (VO table), .asc (ascii)')
    parser_file.add_argument("input", help="input source or file, when prefixed with @ it is assumed to be a text file with a file list (one file per line)")
    parser_file.add_argument("output", help="output file (ends in .hdf5)")
    parser_file.add_argument("columns", help="list of columns to export (or all when empty)", nargs="*")

    args = parser.parse_args(argv[1:])

    verbosity = ["ERROR", "WARNING", "INFO", "DEBUG"]
    logging.getLogger("vaex").setLevel(verbosity[min(3, args.verbose)])
    dataset = None
    if args.task == "soneira":
        if vaex.utils.check_memory_usage(4 * 8 * 2**args.max_level, vaex.utils.confirm_on_console):
            if not args.quiet:
                print("generating soneira peebles dataset...")
            dataset = vaex.file.other.SoneiraPeebles(args.dimension, 2, args.max_level, args.lambdas)
        else:
            return 1
    if args.task == "tap":
        dataset = vaex.dataset.DatasetTap(args.tap_url, args.table_name)
        if not args.quiet:
            print("exporting from {tap_url} table name {table_name} to {output}".format(tap_url=args.tap_url, table_name=args.table_name, output=args.output))
    if args.task == "csv":
        # dataset = vaex.dataset.DatasetTap(args.tap_url, args.table_name)
        if not args.quiet:
            print("exporting from {input} to {output}".format(input=args.input, output=args.output))
    if args.task == "file":
        if args.input[0] == "@":
            inputs = open(args.input[1:]).readlines()
            dataset = vaex.open_many(inputs)
        else:
            dataset = vaex.open(args.input)
        if not args.quiet:
            print("exporting from {input} to {output}".format(input=args.input, output=args.output))

    if dataset is None and args.task not in ["csv"]:
        if not args.quiet:
            print("Cannot open input")
        return 1
    if dataset:
        dataset.set_active_fraction(args.fraction)
    if args.list:
        if not args.quiet:
            print("columns names: " + " ".join(dataset.get_column_names()))
    else:
        if args.task == "csv":
            row_count = -1  # the header does not count
            with file(args.input) as lines:
                for line in lines:
                    row_count += 1
                    # print line
            logger.debug("row_count: %d", row_count)
            with file(args.input) as lines:
                line = next(lines).strip()
                # print line
                names = line.strip().split(",")
                line = next(lines).strip()
                values = line.strip().split(",")
                numerics = []
                for value in values:
                    try:
                        float(value)
                        numerics.append(True)
                    except:
                        numerics.append(False)
                names_numeric = [name for name, numeric in zip(names, numerics) if numeric]
                print(names_numeric)
                output = vaex.file.other.Hdf5MemoryMapped.create(args.output, row_count, names_numeric)
                Ncols = len(names)
                cols = [output.columns[name] if numeric else None for name, numeric in zip(names, numerics)]

                def copy(line, row_index):
                    values = line.strip().split(",")
                    for column_index in range(Ncols):
                        if numerics[column_index]:
                            value = float(values[column_index])
                            cols[column_index][row_index] = value
                row = 0
                copy(line, row)
                row += 1
                progressbar = vaex.utils.progressbar(title="exporting") if args.progress else None
                for line in lines:
                    # print line
                    copy(line, row)
                    row += 1
                    if row % 1000:
                        progressbar.update(row / float(row_count))
                progressbar.finish()
                # print names
        else:
            if args.columns:
                columns = args.columns
            else:
                columns = None
            if columns is None:
                columns = dataset.get_column_names(strings=True, virtual=args.virtual)
            for column in columns:
                if column not in dataset.get_column_names(strings=True, virtual=True):
                    if not args.quiet:
                        print("column %r does not exist, run with --list or -l to list all columns" % column)
                    return 1

            base, output_ext = os.path.splitext(args.output)
            if output_ext not in [".hdf5", ".fits", ".arrow"]:
                if not args.quiet:
                    print("extension %s not supported, only .hdf5, .arrow and .fits are" % output_ext)
                return 1

            if not args.quiet:
                print("exporting %d rows and %d columns" % (len(dataset), len(columns)))
                print("columns: " + " ".join(columns))
            progressbar = vaex.utils.progressbar(title="exporting") if args.progress else None

            def update(p):
                if progressbar:
                    progressbar.update(p)
                return True
            if args.filter:
                dataset.select(args.filter, name='export')
                selection = 'export'
            else:
                selection = None
            if output_ext == ".hdf5":
                export_hdf5(dataset, args.output, column_names=columns, progress=update, shuffle=args.shuffle, sort=args.sort, selection=selection)
            elif output_ext == ".arrow":
                from vaex_arrow.export import export as export_arrow
                export_arrow(dataset, args.output, column_names=columns, progress=update, shuffle=args.shuffle, sort=args.sort, selection=selection)
            elif output_ext == ".fits":
                export_fits(dataset, args.output, column_names=columns, progress=update, shuffle=args.shuffle, sort=args.sort, selection=selection)
            if progressbar:
                progressbar.finish()
            if not args.quiet:
                print("\noutput to %s" % os.path.abspath(args.output))
            dataset.close_files()
    return 0



import ctypes
fcntl = None
try:
    import fcntl
except:
    pass
import threading
from ctypes import *
import os
import ast

import numpy as np
import vaex.column


#libc = cdll.LoadLibrary("libc.dylib")


def getsizeof(ar):
    return ar.nbytes


USE_CACHE = ast.literal_eval(os.environ.get('VAEX_USE_COLUMN_FILE_CACHE', 'False'))
GB = 1024**3
# in GB
CACHE_SIZE =  ast.literal_eval(os.environ.get('VAEX_COLUMN_FILE_CACHE_SIZE', '10'))
if USE_CACHE:
    from cachetools import LRUCache
    cache = LRUCache(maxsize=CACHE_SIZE*GB, getsizeof=getsizeof)
cache_lock = threading.RLock()
F_NOCACHE = 48


class ColumnFile(vaex.column.Column):
    def __init__(self, file, byte_offset, length, dtype, write=False, path=None):
        self.path = path or file.name
        self.file = file
        self.tls = threading.local()
        # keep a record of all duplicate file handles to we can close them
        self.file_handles = []
        self.tls.file = vaex.file.dup(file)
        self.file_handles.append(file)
        self.native = False
        # if hasattr(self.file, 'fileno') and osname
        #     fcntl.fcntl(self.file.fileno(), F_NOCACHE, 1)
            # self.native = True
        #libc.fcntl(self.file.fileno(), fcntl.F_NOCACHE, 1)
        #libc.fcntl(c_int(self.file.fileno()), c_int(fcntl.F_NOCACHE), c_int(1))
        self.byte_offset = byte_offset
        self.length = length
        self.dtype = np.dtype(dtype)
        self.shape = (length,)
        self.write = write

    def __del__(self):
        for f in self.file_handles:
            f.close()

    def __len__(self):
        return self.length

    def trim(self, i1, i2):
        itemsize = self.dtype.itemsize
        byte_offset = self.byte_offset + i1 * itemsize
        length = i2 - i1
        return ColumnFile(self.file, byte_offset, length, self.dtype, self.write, path=self.path)

    def __setitem__(self, slice, values):
        assert self.write, "trying to write to non-writable column"
        start, stop, step = slice.start, slice.stop, slice.step
        start = start or 0
        stop = stop or len(self)
        assert step in [None, 1]
        itemsize = self.dtype.itemsize
        N = stop - start

        page_size = 1024*4
        page_mask = page_size-1
        # TODO: check, it seems tobytes is slow
        ar_bytes = values.tobytes()
        assert len(ar_bytes) == N * itemsize
        # we want to read at a page boundary
        offset = self.byte_offset + start * itemsize

        # make sure we write at a multiple of the page size, if the content is smaller than
        if N * itemsize >= page_size:
            offset_optimal = math.ceil(offset/page_size) * page_size
            padding = offset_optimal - offset
            ar_ptr_pad = ctypes.c_char_p(ar_bytes[:padding])
            ar_ptr_opt = ctypes.c_char_p(ar_bytes[padding:])
            if offset != offset_optimal:
                bytes_write = libc.pwrite(ctypes.c_int32(self.file.fileno()), ar_ptr_pad, ctypes.c_uint64(padding), ctypes.c_uint64(offset))
                if (bytes_write) != padding:
                    raise IOError('write error: expected %d bytes, wrote %d, padding: %d' % (N * itemsize, bytes_write, padding))
            bytes_write = libc.pwrite(ctypes.c_int32(self.file.fileno()), ar_ptr_opt, ctypes.c_uint64(N * itemsize - padding), ctypes.c_uint64(offset_optimal))
            if (bytes_write) != N * itemsize - padding:
                raise IOError('write error: expected %d bytes, wrote %d, padding: %d' % (N * itemsize, bytes_write, padding))
        else:
            ar_ptr = ctypes.c_char_p(ar_bytes)
            bytes_write = libc.pwrite(ctypes.c_int32(self.file.fileno()), ar_ptr, ctypes.c_uint64(N * itemsize), ctypes.c_uint64(offset))
            if (bytes_write) != N * itemsize:
                raise IOError('write error: expected %d bytes, wrote %d, padding: %d' % (N * itemsize, bytes_write, padding))


    def __getitem__(self, slice):
        start, stop, step = slice.start, slice.stop, slice.step
        start = start or 0
        stop = stop or len(self)
        while start < 0:
            start += len(self)
        while stop < 0:
            stop += len(self)
        assert step in [None, 1]
        itemsize = self.dtype.itemsize
        items = stop - start
        key = (self.path, self.byte_offset, start, stop)
        if USE_CACHE:
            with cache_lock:
                ar = cache.get(key)
        else:
            ar = None
        if ar is None:
            N = items
            offset = self.byte_offset + start * itemsize
            if self.native:
                # fcntl.fcntl(self.file.fileno(), fcntl.F_NOCACHE, 1)
                page_size = 1024*4
                page_mask = page_size-1
                ar_bytes = np.zeros(N*itemsize + page_size, np.uint8)
                ar_ptr = ar_bytes.ctypes.data_as(ctypes.POINTER(ctypes.c_byte))
                # we want to read at a page boundary
                offset_optimal = offset & ~page_mask
                padding = offset - offset_optimal
                bytes_read = libc.pread(ctypes.c_int32(self.file.fileno()), ar_ptr, ctypes.c_uint64(N * itemsize + padding), ctypes.c_uint64(offset_optimal))
                if (bytes_read-padding) != N * itemsize:
                    raise IOError('read error: expected %d bytes, read %d, padding: %d' % (N * itemsize, bytes_read, padding))
                ar = np.frombuffer(ar_bytes, self.dtype, offset=padding, count=N)
            else:
                byte_length = items*itemsize
                offset = self.byte_offset + start * itemsize
                # Quick and safe way to get the thread local file handle:
                file = getattr(self.tls, 'file', None)
                if file is None:
                    with cache_lock:
                        file = getattr(self.tls, 'file', None)
                        if file is None:
                            file = self.tls.file = vaex.file.dup(self.file)
                            self.file_handles.append(file)
                # this is the fast path, that avoids a memory copy but gets a view on the underlying data
                # cache.py:CachedFile supports this
                if hasattr(file, '_as_numpy'):
                    ar = file._as_numpy(offset, byte_length, self.dtype)
                else:
                    # Traditinal file object go this slower route
                    # and they need per thread file object since the location (seek)
                    # is in the state of the file object

                    file.seek(offset)
                    data = file.read(byte_length)
                    ar = np.frombuffer(data, self.dtype, count=N)
            if USE_CACHE:
                with cache_lock:
                    cache[key] = ar
        return ar



from __future__ import absolute_import
__author__ = 'maartenbreddels'

import tornado.ioloop
import tornado.web
import tornado.httpserver
import tornado.websocket
import tornado.auth
import tornado.gen
import threading
import logging
import vaex as vx
import vaex.utils
import json
import inspect
import yaml
import argparse
import os
import time
import numpy as np
import vaex.events
import collections
import concurrent.futures
import vaex.execution
import vaex.multithreading
import tornado.escape
from cachetools import Cache, LRUCache
import sys

logger = logging.getLogger("vaex.webserver")
job_index = 0


class JobFlexible(object):
    def __init__(self, cost, fn=None, args=None, kwargs=None, index=None):
        global job_index
        self.cost = cost
        if index is None:
            self.index = job_index
            job_index += 1
        else:
            self.index = index
        self.fraction = None
        self.time_elapsed_goal = None
        self.time_created = time.time()
        self.fn = fn or (lambda *args, **kwargs: None)
        self.args = args
        self.kwargs = kwargs

    def set_fraction(self, fraction):
        self.fraction = fraction

    def set_time_elapsed_goal(self, time_elapsed_goal):
        self.time_elapsed_goal = time_elapsed_goal

    def execute(self):
        try:
            result = self.fn(*self.args, **self.kwargs)
            return result
        finally:
            self.done()

    def done(self):
        self.time_done = time.time()
        self.time_elapsed = self.time_done - self.time_created


class JobQueue(object):
    def __init__(self):
        self.queue = []
        self.queue_delay = 1.  # 1 second to get rid of the queue
        self.seconds_per_cost = 1.
        self.cost_modifier_per_miss = 0.1
        self.fraction_minimum = 1e-4
        self.signal_job_finished = vaex.events.Signal("job finished")

    def add(self, job):
        self.queue.append(job)

    def get_next(self):
        if len(self.queue) == 0:
            raise IndexError("queue empty")
        total_cost = sum(job.cost for job in self.queue)
        job_count = len(self.queue)
        job = self.queue.pop(0)
        logger.debug("queue length is %d, total cost of queue is %f, job index %d", len(self.queue), total_cost, job.index)
        # if the total cost equals seconds, and we want the queue to be finished in 1 second
        # this will be the best fraction
        elapsed = time.time() - job.time_created
        if elapsed > self.queue_delay * 1.3:  # we already spent way too long time, the math doesn't work out, so what now?
            # we scale the cost up, since it apearently doesn't work out fully
            # raise ValueError, "elapsed = %f" % elapsed
            logger.error("queue clogging up")
        # assuming the queue costs will stay constant we want to finished this whole queue in max self.queue_delay
        # seconds. Therefore the current job should only calculate a fraction as computed by:
        fraction = (self.queue_delay - elapsed) / (total_cost * self.seconds_per_cost) * 0.9
        fraction = min(max(self.fraction_minimum, fraction), 1)
        time_elapsed_goal = elapsed + job.cost * self.seconds_per_cost * fraction / 0.9  # elapsed + fraction * (self.seconds_per_cost)
        logger.debug("to finish the queue we set the next job's fraction to %f (time elapsed = %f)", fraction, elapsed)
        # print fraction
        job.set_fraction(fraction)
        job.set_time_elapsed_goal(time_elapsed_goal)
        return job

    def finished(self, job):
        ratio = job.time_elapsed / job.time_elapsed_goal  # / job.time_elapsed_goal
        if ratio > 1.2:
            logger.debug("job should have taken %f seconds, took %f seconds, we should go faster", job.time_elapsed_goal, job.time_elapsed)
            self.seconds_per_cost *= (1 + self.cost_modifier_per_miss)
        if ratio < 0.75:
            self.seconds_per_cost *= (1 - self.cost_modifier_per_miss)
            logger.debug("job should have taken %f seconds, took %f seconds, we can slow down", job.time_elapsed_goal, job.time_elapsed)
        logger.debug("seconds_per_cost set to: %f", self.seconds_per_cost)
        self.signal_job_finished.emit(job)

        # seconds_goal = job.set_fraction
        # fraction_goal = job.set_fraction
        # fraction_achieved = job.ti
        pass

# import concurrent.future


class JobExecutor(object):
    def __init__(self, job_queue):
        self.job_queue = job_queue
        # self.thread_pool_executor = concurrent.

    def empty_queue(self):
        while True:
            try:
                job = self.job_queue.get_next()
                job.execute()
                self.job_queue.finished(job)
            except IndexError:
                return


class GoogleOAuth2LoginHandler(tornado.web.RequestHandler,
                               tornado.auth.GoogleOAuth2Mixin):
    @tornado.gen.coroutine
    def get(self):
        if self.get_argument('code', False):
            user = yield self.get_authenticated_user(
                redirect_uri='http://your.site.com/auth/google',
                code=self.get_argument('code'))
            # Save the user with e.g. set_secure_cookie
        else:
            yield self.authorize_redirect(
                redirect_uri='http://your.site.com/auth/google',
                client_id='cid',
                scope=['profile', 'email'],
                response_type='code',
                extra_params={'approval_prompt': 'auto'})


def task_invoke(subspace, method_name, **arguments):
    method = getattr(subspace, method_name)
    # args, varargs, kwargs, defaults = inspect.getargspec(method)
    # print inspect.getargspec(method)
    # args_required = args[:len(args)-len(defaults)]
    # kwargs = {}
    # for arg in args:
    # if arg in arguments:
    # kwargs[arg] = arguments[arg]
    # print("!!!!!! calling %s with arguments %r" % (method_name, arguments))
    values = method(**arguments)
    return values


class ListHandler(tornado.web.RequestHandler):
    def initialize(self, webserver, submit_threaded, cache, cache_selection, datasets=None):
        self.webserver = webserver
        self.submit_threaded = submit_threaded
        self.cache = cache
        self.cache_selection = cache_selection

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")

    @tornado.gen.coroutine
    def get(self):
        user_id = None
        if False:  # disable user_id for the moment
            user_id = self.get_cookie("user_id")
            logger.debug("user_id in cookie: %r", user_id)
            if user_id is None:
                import uuid
                user_id = str(uuid.uuid4())
                self.set_cookie("user_id", user_id)
            logger.debug("user_id: %r", user_id)
        # tornado retursn a list of values, just use the first value
        arguments = {key: json.loads(tornado.escape.to_basestring(self.request.arguments[key][0])) for key in self.request.arguments.keys()}
        key = (self.request.path, "-".join([str(v) for v in sorted(arguments.items())]))
        response = self.cache.get(key)
        if response is None:
            response = yield self.submit_threaded(process, self.webserver, user_id, self.request.path, **arguments)
            try:
                self.cache[key] = response
                pass
            except ValueError:
                pass  # raised when it doesn't fit in cache
        else:
            logger.debug("cache hit for key: %r", key)
        # logger.debug("response is: %r", response)
        if response is None:
            response = self.error("unknown request or error")
        if isinstance(response, (np.ndarray, float)):
            response = np.array(response)
            self.set_header("Content-Type", "application/numpy-array")
            self.write(str(response.shape) + "\n")
            self.write(str(response.dtype) + "\n")
            self.write(response.tobytes())
        else:
            if isinstance(response, str):
                self.set_header("Content-Type", "application/octet-stream")
            self.write(response)


class ProgressWebSocket(tornado.websocket.WebSocketHandler):
    def initialize(self, webserver, submit_threaded, cache, cache_selection, datasets=None):
        self.webserver = webserver
        self.submit_threaded = submit_threaded
        self.cache = cache
        self.cache_selection = cache_selection

    def check_origin(self, origin):
        return True

    def open(self):
        logger.debug("WebSocket opened")

    @tornado.gen.coroutine
    def on_message(self, message):
        logger.debug("websocket message: %r", message)
        arguments = json.loads(message)
        path = arguments["path"]
        arguments.pop("path")
        user_id = arguments.pop("user_id", None)
        job_id = arguments.pop("job_id")

        if False:  # disable user_id for the moment
            user_id = self.get_cookie("user_id")
            if user_id is None:
                user_id = arguments.pop("user_id")
                if user_id is None:
                    self.write_json(error="no user id, do an http get request first")
                    return
        last_progress = [None]

        def progress(f):
            def do():
                logger.debug("progress: %r", f)
                last_progress[0] = f
                self.write_json(job_id=job_id, job_phase="PENDING", progress=f)
            if last_progress[0] is None or (f - last_progress[0]) > 0.05 or f == 1.0:
                self.webserver.ioloop.add_callback(do)
            return True
        key = (path, "-".join([str(v) for v in sorted(arguments.items())]))
        response = self.cache.get(key)
        progress(0)
        if response is None:
            # response = yield self.submit_threaded(process, self.webserver, user_id, self.request.path, **arguments)
            response = yield self.submit_threaded(process, self.webserver, user_id, path, progress=progress,
                                                  **arguments)
            try:
                self.cache[key] = response
                pass
            except ValueError:
                pass  # raised when it doesn't fit in cache
        progress(1)

        if response is None:
            response = self.error("unknown request or error")

        meta_data = dict(job_id=job_id)
        if isinstance(response, (np.ndarray, float)):
            response = np.array(response)
            # self.set_header("Content-Type", "application/numpy-array")
            meta_data["shape"] = str(response.shape)
            meta_data["dtype"] = str(response.dtype)
            meta_data["job_phase"] = "COMPLETED"
            import zlib
            data_response = zlib.compress(response.tobytes())
        else:  # it should be a dict
            if "result" in response:
                response["job_phase"] = "COMPLETED"
            if "error" in response:
                response["job_phase"] = "ERROR"
            if "exception" in response:
                response["job_phase"] = "EXCEPTION"
            meta_data.update(response)
            data_response = b""
        try:
            text_response = json.dumps(meta_data).encode("utf8")\
                + b"\n" + data_response
        except Exception as e:
            meta_data = dict(job_id=job_id, job_phase='EXCEPTION', **exception(e))
            text_response = json.dumps(meta_data).encode("utf8")\
                + b"\n" + data_response

        self.write_message(text_response, binary=True)

    def write_json(self, **kwargs):
        logger.debug("writing json: %r", kwargs)
        self.write_message(json.dumps(kwargs) + "\n", binary=True)

    def on_close(self):
        logger.debug("WebSocket closed")


class QueueHandler(tornado.web.RequestHandler):
    def initialize(self, datasets):
        self.datasets = datasets
        self.datasets_map = dict([(ds.name, ds) for ds in self.datasets])

    def get(self):
        # self.write("Hello, world")
        self.write(dict(datasets=[{"name": ds.name, "length": len(ds)} for ds in self.datasets]))


def exception(exception):
    logger.exception("handled exception at server, all fine")
    return ({"exception": {"class": str(exception.__class__.__name__), "msg": str(exception)}})


def error(msg):
    return ({"error": msg})  # , "result":None})


def process(webserver, user_id, path, fraction=None, progress=None, **arguments):
    if not hasattr(webserver.thread_local, "executor"):
        logger.debug("creating thread pool and executor")
        webserver.thread_local.thread_pool = vaex.multithreading.ThreadPoolIndex(max_workers=webserver.threads_per_job)
        webserver.thread_local.executor = vaex.execution.Executor(thread_pool=webserver.thread_local.thread_pool)
        webserver.thread_pools.append(webserver.thread_local.thread_pool)

    progress = progress or (lambda x: True)
    progress(0)
    # TODO: mem leak and other issues if we don't disconnect this
    webserver.thread_local.executor.signal_progress.connect(progress)
    # return ("Hello, world")
    # print request.path
    try:
        parts = [part for part in path.split("/") if part]
        logger.debug("request: %r" % parts)
        if parts[0] == "sleep":
            seconds = float(parts[1])
            import time
            time.sleep(min(1, seconds))
            return ({"result": seconds})
        elif parts[0] == "datasets":
            if len(parts) == 1:
                #
                response = dict(result=[{'name': ds.name,
                                         'length_original': ds.length_original(),
                                         'column_names': ds.get_column_names(strings=True),
                                         'dtypes': {name: str("str" if ds.dtype(name) == vaex.column.str_type else ds.dtype(name)) for name in ds.get_column_names(strings=True)},
                                         'state': ds.state_get()
                                         } for ds in webserver.datasets])
                logger.debug("response: %r", response)
                return response
            else:

                dataset_name = parts[1]
                logger.debug("dataset: %s", dataset_name)
                if dataset_name not in webserver.datasets_map:
                    error("dataset does not exist: %r, possible options: %r" % (dataset_name, webserver.datasets_map.keys()))
                else:
                    if len(parts) > 2:
                        method_name = parts[2]
                        if "expressions" in arguments:
                            expressions = arguments["expressions"]
                        else:
                            expressions = None
                        # make a shallow copy, such that selection and active_fraction is not shared
                        dataset_original = webserver.datasets_map[dataset_name]
                        dataset = dataset_original.copy()
                        # copy the cache, 2 level deep clone
                        # currently not working, could be a good optimization
                        # dataset._selection_mask_caches = {key:dict(value) for key, value in dataset_original._selection_mask_caches.items()}
                        logger.info("method: %r args: %r" % (method_name, arguments))
                        # TODO: executor should be an argument to the stats functions..
                        dataset.executor = webserver.thread_local.executor
                        if dataset.mask is not None:
                            logger.debug("selection: %r", dataset.mask.sum())
                        if 'state' in arguments:
                            dataset.state_set(arguments['state'], use_active_range=True)
                        if expressions:
                            for expression in expressions:
                                try:
                                    dataset.validate_expression(expression)
                                except (SyntaxError, KeyError, NameError) as e:
                                    return exception(e)
                        subspace = dataset(*expressions, executor=webserver.thread_local.executor) if expressions else None
                        if dataset_original.has_selection(vaex.dataset.FILTER_SELECTION_NAME):
                            selection_original = dataset_original.get_selection(vaex.dataset.FILTER_SELECTION_NAME)
                            if not dataset.has_selection(vaex.dataset.FILTER_SELECTION_NAME):
                                dataset.set_selection(selection_original, name=vaex.dataset.FILTER_SELECTION_NAME)
                            else:
                                selection = dataset.get_selection(vaex.dataset.FILTER_SELECTION_NAME)
                                # UGLY: we 'stitch' the current selection at the end of the
                                # selection 'list'
                                selection_end = selection
                                while selection_end.previous_selection:
                                    selection_end = selection_end.previous_selection
                                selection_end.previous_selection = selection_original
                                selection_end.mode = "and"
                        try:
                            if subspace:
                                if 'selection' in arguments and arguments['selection']:
                                    subspace = subspace.selected()
                            if subspace is None:
                                for name in "job_id state auto_fraction expressions active_fraction selections variables virtual_columns active_start_index active_end_index".split():
                                    arguments.pop(name, None)
                            else:
                                if subspace is not None:
                                    for name in "job_id state auto_fraction expressions active_fraction selection selections variables virtual_columns active_start_index active_end_index".split():
                                        arguments.pop(name, None)
                            logger.debug("subspace: %r", subspace)
                            if subspace is None and method_name in "count cov correlation covariance mean std minmax min max sum var".split():
                                grid = task_invoke(dataset, method_name, **arguments)
                                return grid
                            elif method_name in ["minmax", "image_rgba_url", "var", "mean", "sum", "limits_sigma", "nearest", "correlation", "mutual_information"]:
                                # print "expressions", expressions
                                values = task_invoke(subspace, method_name, **arguments)
                                logger.debug("result: %r", values)
                                # print values, expressions
                                values = values.tolist() if hasattr(values, "tolist") else values

                                # for value in values:
                                # logger.debug("value: %r" % type(value))
                                # try:
                                # values = [(k.tolist() if hasattr(k, "tolist") else k) for k in values]
                                # except:
                                # pass
                                return ({"result": values})
                            elif method_name == "histogram":
                                grid = task_invoke(subspace, method_name, **arguments)
                                return grid
                                # self.set_header("Content-Type", "application/octet-stream")
                                # return (grid.tostring())
                                # import base64
                                # return grid.tostring()
                                # result = base64.b64encode(grid.tostring()).decode("ascii")
                                # logger.debug("type: %s" % type(result))
                                # return {"result": result}
                            elif method_name in ["select", "lasso_select"]:
                                dataset.mask = webserver.cache_selection.get((dataset.path, user_id))
                                result = task_invoke(dataset, method_name, **arguments)
                                result = result.tolist() if hasattr(result, "tolist") else result
                                webserver.cache_selection[(dataset.path, user_id)] = dataset.mask
                                return ({"result": result})
                            elif method_name in ["evaluate"]:
                                result = task_invoke(dataset, method_name, **arguments)
                                # evaluating a string results in dtype = object, but that is difficult
                                # to (de)serialize, better would be to serialize the arrow arrays
                                if dataset.dtype(arguments['expression']) == vaex.column.str_type:
                                    result = result.astype(vaex.column.str_type)
                                return result
                            else:
                                logger.error("unknown method: %r", method_name)
                                return error("unknown method: " + method_name)
                        except (SyntaxError, KeyError, NameError) as e:
                            return exception(e)
    except Exception as e:
        logger.exception("unknown issue")
        return exception(e)
    return error("unknown request")


MB = 1024**2
GB = MB * 1024


class WebServer(threading.Thread):
    def __init__(self, address="localhost", port=9000, webserver_thread_count=2, cache_byte_size=500 * MB,
                 cache_selection_byte_size=500 * MB, datasets=[], compress=True, development=False, threads_per_job=4):
        threading.Thread.__init__(self)
        self.setDaemon(True)
        self.address = address
        self.port = port
        self.started = threading.Event()
        self.set_datasets(datasets)

        self.webserver_thread_count = webserver_thread_count
        self.threads_per_job = threads_per_job

        self.thread_pool = concurrent.futures.ThreadPoolExecutor(self.webserver_thread_count)
        self.thread_local = threading.local()
        self.thread_pools = []

        self.job_queue = JobQueue()

        self.cache = LRUCache(cache_byte_size, getsizeof=sys.getsizeof)
        self.cache_selection = LRUCache(cache_selection_byte_size, getsizeof=sys.getsizeof)

        self.options = dict(webserver=self, datasets=datasets, submit_threaded=self.submit_threaded, cache=self.cache,
                            cache_selection=self.cache_selection)

        # tornado.web.GZipContentEncoding.MIN_LENGTH = 1
        tornado.web.GZipContentEncoding.CONTENT_TYPES.add("application/octet-stream")
        self.application = tornado.web.Application([
            (r"/queue", QueueHandler, self.options),
            (r"/auth", GoogleOAuth2LoginHandler, {}),
            (r"/websocket", ProgressWebSocket, self.options),
            (r"/.*", ListHandler, self.options),
        ], compress_response=compress, debug=development)
        logger.debug("compression set to %r", compress)
        logger.debug("cache size set to %s", vaex.utils.filesize_format(cache_byte_size))
        logger.debug("thread count set to %r", self.webserver_thread_count)

    def set_datasets(self, datasets):
        self.datasets = list(datasets)
        self.datasets_map = dict([(ds.name, ds) for ds in self.datasets])

    def submit_threaded(self, callable, *args, **kwargs):
        job = JobFlexible(4., callable, args=args, kwargs=kwargs)
        self.job_queue.add(job)

        def execute():
            job = self.job_queue.get_next()
            job.kwargs["fraction"] = job.fraction  # add fraction keyword argument to the callback
            try:
                result = job.execute()
            finally:
                self.job_queue.finished(job)
            return result
        future = self.thread_pool.submit(execute)  # , *args, **kwargs)
        return future

    def serve(self):
        self.mainloop()

    def serve_threaded(self):
        logger.debug("start thread")
        asyncio = None
        try:
            import asyncio
        except:
            pass
        if asyncio and tornado.version_info[0] >= 5:
            from tornado.platform.asyncio import AnyThreadEventLoopPolicy
            # see https://github.com/tornadoweb/tornado/issues/2308
            asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
        self.start()
        logger.debug("wait for thread to run")
        self.started.wait()
        logger.debug("make tornado io loop the main thread's current")
        # this will make the main thread use this ioloop as current
        self.ioloop.make_current()

    def run(self):
        self.mainloop()

    def mainloop(self):
        logger.info("serving at http://%s:%d" % (self.address, self.port))
        self.ioloop = tornado.ioloop.IOLoop.current()
        # listen doesn't return a server object, which we need to close
        # self.application.listen(self.port, address=self.address)
        from tornado.httpserver import HTTPServer
        self.server = HTTPServer(self.application)
        try:
            self.server.listen(self.port, self.address)
        except:
            self.started.set()
            raise
        self.started.set()
        # self.ioloop.add_callback(self.started.set)
        # if not self.ioloop.is_running():
        asyncio = None
        try:
            import asyncio
        except:
            pass
        if asyncio and tornado.version_info[0] >= 5:
            from tornado.platform.asyncio import AnyThreadEventLoopPolicy
            # see https://github.com/tornadoweb/tornado/issues/2308
            asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())

        try:
            self.ioloop.start()
        except RuntimeError:
            pass  # TODO: not sure why this happens in the unittest
        # self.ioloop.stop()
        # try:
        # self.ioloop.close()
        # except ValueError:
        # pass
        # self.ioloop.clear_current()

    def stop_serving(self):
        logger.debug("stop server")
        self.server.stop()
        logger.debug("stop io loop")
        self.ioloop.stop()
        # self.ioloop.close()
        # self.ioloop.clear_current()
        for thread_pool in self.thread_pools:
            thread_pool.shutdown()


defaults_yaml = """
address: 0.0.0.0
port: 9000
filenames: []
verbose: 2
cache: 500000000
compress: true
filename: []
development: False
threads_per_job: 4
"""


def main(argv):

    parser = argparse.ArgumentParser(argv[0])
    parser.add_argument("filename", help="filename for dataset", nargs='*')
    parser.add_argument("--address", help="address to bind the server to (default: %(default)s)", default="0.0.0.0")
    parser.add_argument("--port", help="port to listen on (default: %(default)s)", type=int, default=9000)
    parser.add_argument('--verbose', '-v', action='count', default=2)
    parser.add_argument('--cache', help="cache size in bytes for requests, set to zero to disable (default: %(default)s)", type=int, default=500000000)
    parser.add_argument('--compress', help="compress larger replies (default: %(default)s)", default=True, action='store_true')
    parser.add_argument('--no-compress', dest="compress", action='store_false')
    parser.add_argument('--development', default=False, action='store_true', help="enable development features (auto reloading)")
    parser.add_argument('--threads-per-job', default=4, type=int, help="threads per job (default: %(default)s)")
    # config = layeredconfig.LayeredConfig(defaults, env, layeredconfig.Commandline(parser=parser, commandline=argv[1:]))
    config = parser.parse_args(argv[1:])

    verbosity = ["ERROR", "WARNING", "INFO", "DEBUG"]
    logging.getLogger("vaex").setLevel(verbosity[config.verbose])
    # import vaex
    # vaex.set_log_level_debug()
    from vaex.settings import webserver as settings

    # filenames = config.filenames
    filenames = []
    filenames = config.filename
    datasets = []
    for filename in filenames:
        ds = vx.open(filename)
        if ds is None:
            print("error opening file: %r" % filename)
        else:
            datasets.append(ds)
    datasets = datasets or [vx.example()]
    # datasets = [ds for ds in datasets if ds is not None]
    logger.info("datasets:")
    for dataset in datasets:
        logger.info("\thttp://%s:%d/%s or ws://%s:%d/%s", config.address, config.port, dataset.name, config.address, config.port, dataset.name)
    server = WebServer(datasets=datasets, address=config.address, port=config.port, cache_byte_size=config.cache,
                       compress=config.compress, development=config.development,
                       threads_per_job=config.threads_per_job)
    server.serve()


if __name__ == "__main__":
    # logger.setLevel(logging.logging.DEBUG)
    # print sys.argv
    main(sys.argv)
    # 3_threaded()
    # import time
    # time.sleep(10)
    # server.stop_serving()



__author__ = 'breddels'
# due to 32 bit limitations in numpy, we cannot use astropy's fits module for writing colfits
import sys
import math

import vaex.dataset
import astropy.io.fits
import numpy as np
import logging

logger = logging.getLogger("vaex.file.colfits")

def empty(filename, length, column_names, data_types, data_shapes, ucds, units, null_values={}):
	with open(filename, "wb") as f:
		logger.debug("preparing empty fits file: %s", filename)
		class Scope(object):
			pass

		def write(key, value, comment=""):
			first_part = "{key:8}= {value:20} / ".format(key=key, value=value)
			f.write(first_part.encode("ascii"))
			leftover = 80 - len(first_part)
			f.write(("{comment:"+str(leftover) +"}").format(comment=comment).encode("ascii"))
			logger.debug("at pos: %s", f.tell())
		def finish_header():
			f.write("{end:80}".format(end="END").encode("ascii"))
			offset = f.tell()
			bytes_over_padding = offset % 2880
			logger.debug(("bytes_over_padding: %s", bytes_over_padding))
			if bytes_over_padding > 0:
				padding = 2880 - bytes_over_padding
				f.write((" "*padding).encode("ascii"))
		def finish_data():
			offset = f.tell()
			bytes_over_padding = offset % 2880
			if bytes_over_padding > 0:
				padding = 2880 - bytes_over_padding
				f.write(("\0"*padding).encode("ascii"))

		byte_size = sum([length * type.itemsize for type in data_types])

		write("SIMPLE", "T", "file conforms to FITS standard")
		write("BITPIX", 8, "number of bits per data pixel")
		write("NAXIS", 0, "number of array dimensions")
		finish_header()
		write("XTENSION", repr("BINTABLE"), "binary table extension")

		write("BITPIX", 8, "number of bits per data pixel")
		write("NAXIS", 2, "number of array dimensions")
		write("NAXIS1", byte_size, "length of dim 1")
		write("NAXIS2", 1, "length of dim 2")
		write("PCOUNT", 0, "number of group parameters")
		write("GCOUNT", 1, "number of groups")

		write("TFIELDS", len(column_names), "number of columns")
		for i, (column_name, type, shape) in enumerate(zip(column_names, data_types, data_shapes)):
			i += 1 # 1 based index
			#column = dataset.columns[column_name]
			write("TTYPE%d" % i, repr(str(column_name)), "column name %i" % (i))
			numpy_type_name = type.descr[0][1][1:] # i4, f8 etc
			if numpy_type_name[0] == 'S':
				string_length = numpy_type_name[1:]
				fits_type = str(int(string_length)*length)+"A"
				logger.debug("type for %s: numpy=%r, fits=%r, string_length=%r length=%r", column_name, numpy_type_name, fits_type, string_length, length)
				# TODO: support rank1 arrays
				write("TFORM%d" % i , repr("{type}".format(type=fits_type)), "")
				write("TDIM%d" % i, repr("({string_length},{length})".format(string_length=string_length, length=length)), "")
			else:
				fits_type = astropy.io.fits.column.NUMPY2FITS[numpy_type_name]
				logger.debug("type for %s: numpy=%r, fits=%r", column_name, numpy_type_name, fits_type)
				# TODO: support rank1 arrays
				write("TFORM%d" % i , repr("{length}{type}".format(length=length, type=fits_type)), "")
				write("TDIM%d" % i, repr("(1,{length})".format(length=length)), "")
			ucd = ucds[i-1]
			if ucd:
				write("TUCD%d" % i, repr(str(ucd)))
			unit = units[i-1]
			if unit:
				write("TUNIT%d" % i, repr(str(unit)))
			if column_name in null_values:
				write("TNULL%d" % i, str(null_values[column_name]))

		finish_header()

		for i, (column_name, type, shape) in enumerate(zip(column_names, data_types, data_shapes)):
			byte_size = length * type.itemsize
			f.seek(f.tell() + byte_size)
		finish_data()

def write_colfits(dataset, path, selection=False):
	with open(path, "wb") as f:
		class Scope(object):
			pass

		vars = Scope()
		#vars.
		def write(key, value, comment=""):
			f.write("{key:8}= {value:20} / {comment:47}".format(key=key, value=value, comment=comment))
			print(("at pos", f.tell()))
		def finish_header():
			print((f.write("{end:80}".format(end="END"))))
			offset = f.tell()
			bytes_over_padding = offset % 2880
			print(("bytes_over_padding", bytes_over_padding))
			if bytes_over_padding > 0:
				padding = 2880 - bytes_over_padding
				f.write(" "*padding)
		def finish_data():
			offset = f.tell()
			bytes_over_padding = offset % 2880
			if bytes_over_padding > 0:
				padding = 2880 - bytes_over_padding
				f.write("\0"*padding)


		write("SIMPLE", "T", "file conforms to FITS standard")
		write("BITPIX", 8, "number of bits per data pixel")
		write("NAXIS", 0, "number of array dimensions")
		finish_header()
		write("XTENSION", repr("BINTABLE"), "binary table extension")

		write("BITPIX", 8, "number of bits per data pixel")
		write("NAXIS", 2, "number of array dimensions")
		write("NAXIS1", dataset.byte_size(selection=selection), "length of dim 1")
		write("NAXIS2", 1, "length of dim 2")

		write("TFIELDS", len(dataset.column_names), "number of columns")
		for i, column_name in enumerate(dataset.column_names):
			i += 1 # 1 based index
			column = dataset.columns[column_name]
			write("TTYPE%d" % i, repr(str(column_name)), "column name %i" % (i))
			numpy_type_name = column.dtype.descr[0][1][1:] # i4, f8 etc
			fits_type = astropy.io.fits.column.NUMPY2FITS[numpy_type_name]
			# TODO: support rank1 arrays
			write("TFORM%d" % i , repr("{length}{type}".format(length=len(dataset), type=fits_type)), "")
			write("TDIM%d" % i, repr("(1,{length})".format(length=len(dataset))), "")

		finish_header()

		for i, column_name in enumerate(dataset.column_names):
			column = dataset.columns[column_name]
			numpy_type_name = column.dtype.descr[0][1][1:] # i4, f8 etc
			fits_type = astropy.io.fits.column.NUMPY2FITS[numpy_type_name]
			chunk_size = 1024**2 # 1 meg at a time
			chunks = int(math.ceil(len(dataset)/float(chunk_size)))
			for i in range(chunks):
				i1 = i * chunk_size
				i2 = min(len(dataset), (i+1) * chunk_size)
				data_big_endian = column[i1:i2].astype(">" + numpy_type_name)
				f.write(data_big_endian)
			print((f.tell(), f.tell() / 1024**2, "mb", len(dataset)))
			assert i2 == len(dataset)
		finish_data()



if __name__ == "__main__":
	input = sys.argv[1]
	output = sys.argv[2]
	dataset_in = vaex.dataset.load_file(input)
	write_colfits(dataset_in, output)


try:
    from urllib.parse import urlparse, parse_qs
except ImportError:
    from urlparse import urlparse, parse_qs

try:
    import s3fs
except Exception as e:
    import_exception = e
    s3fs = None

import vaex.file.cache


normal_open = open


def is_s3_path(path):
    return path.startswith('s3://')

def dup(f):
    return f.s3.open(f.path, f.mode)

def open(path, mode='rb', **kwargs):
    if not is_s3_path(path):
        return normal_open(path, mode)
    if s3fs is None:
        raise import_exception
    o = urlparse(path)
    assert o.scheme == 's3'
    naked_path = path
    if '?' in naked_path:
        naked_path = naked_path[:naked_path.index('?')]
    # only use the first item
    options = {key: values[0] for key, values in parse_qs(o.query).items()}
    options.update(kwargs)
    use_cache = options.get('cache', 'true') in ['true', 'True', '1']
    if 'cache' in options:
        del options['cache']
    anon = options.get('anon', 'false') in ['true', 'True', '1']
    if 'anon' in options:
        del options['anon']
    s3 = s3fs.S3FileSystem(anon=anon, default_block_size=1,
                           default_fill_cache=False, **options)
    if vaex.utils.should_cache() and use_cache:
        fp = lambda: s3.open(naked_path, mode)
        fp = vaex.file.cache.CachedFile(fp, naked_path)
    else:
        fp = s3.open(naked_path, mode)
    return fp



from .superstrings import *
import numpy as np


def array(ar):
    if isinstance(ar, (tuple, list)):
        ar = np.asarray(ar)
    if isinstance(ar, np.ndarray):
        return StringArray(ar.astype('O'))
    elif isinstance(ar, (StringList32, StringList64)):
        return ar
    else:
        raise ValueError('Cannot convert %r to a string array' % ar)


import base64
import cloudpickle as pickle
import functools
import operator
import six
import collections

from future.utils import with_metaclass
import numpy as np
import tabulate

from vaex.utils import _ensure_strings_from_expressions, _ensure_string_from_expression
from vaex.column import ColumnString, _to_string_sequence, str_type
from .hash import counter_type_from_dtype
import vaex.serialize
from . import expresso


try:
    from StringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO

try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections


# TODO: repeated from dataframe.py
default_shape = 128
PRINT_MAX_COUNT = 10

expression_namespace = {}
expression_namespace['nan'] = np.nan


expression_namespace = {}
expression_namespace['nan'] = np.nan


_binary_ops = [
    dict(code="+", name='add', op=operator.add),
    dict(code="in", name='contains', op=operator.contains),
    dict(code="/", name='truediv', op=operator.truediv),
    dict(code="//", name='floordiv', op=operator.floordiv),
    dict(code="&", name='and', op=operator.and_),
    dict(code="^", name='xor', op=operator.xor),

    dict(code="|", name='or', op=operator.or_),
    dict(code="**", name='pow', op=operator.pow),
    dict(code="is", name='is', op=operator.is_),
    dict(code="is not", name='is_not', op=operator.is_not),

    dict(code="<<", name='lshift', op=operator.lshift),
    dict(code="%", name='mod', op=operator.mod),
    dict(code="*", name='mul', op=operator.mul),

    dict(code=">>", name='rshift', op=operator.rshift),
    dict(code="-", name='sub', op=operator.sub),

    dict(code="<", name='lt', op=operator.lt),
    dict(code="<=", name='le', op=operator.le),
    dict(code="==", name='eq', op=operator.eq),
    dict(code="!=", name='ne', op=operator.ne),
    dict(code=">=", name='ge', op=operator.ge),
    dict(code=">", name='gt', op=operator.gt),
]
if hasattr(operator, 'div'):
    _binary_ops.append(dict(code="/", name='div', op=operator.div))
if hasattr(operator, 'matmul'):
    _binary_ops.append(dict(code="@", name='matmul', op=operator.matmul))

reversable = 'add sub mul matmul truediv floordiv mod divmod pow lshift rshift and xor or'.split()

_unary_ops = [
    dict(code="~", name='invert', op=operator.invert),
    dict(code="-", name='neg', op=operator.neg),
    dict(code="+", name='pos', op=operator.pos),
]


class Meta(type):
    def __new__(upperattr_metaclass, future_class_name,
                future_class_parents, attrs):
        # attrs = {}
        for op in _binary_ops:
            def wrap(op=op):
                def f(a, b):
                    self = a
                    # print(op, a, b)
                    if isinstance(b, Expression):
                        assert b.ds == a.ds
                        b = b.expression
                    elif isinstance(b, (np.timedelta64)):
                        df = a.ds
                        b = df.add_variable('var_time_delta', b, unique=True)
                    expression = '({0} {1} {2})'.format(a.expression, op['code'], b)
                    return Expression(self.ds, expression=expression)
                attrs['__%s__' % op['name']] = f
                if op['name'] in reversable:
                    def f(a, b):
                        self = a
                        # print(op, a, b)
                        if isinstance(b, Expression):
                            assert b.ds == a.ds
                            b = b.expression
                        expression = '({2} {1} {0})'.format(a.expression, op['code'], b)
                        return Expression(self.ds, expression=expression)
                    attrs['__r%s__' % op['name']] = f

            wrap(op)
        for op in _unary_ops:
            def wrap(op=op):
                def f(a):
                    self = a
                    expression = '{0}({1})'.format(op['code'], a.expression)
                    return Expression(self.ds, expression=expression)
                attrs['__%s__' % op['name']] = f
            wrap(op)
        return type(future_class_name, future_class_parents, attrs)


class DateTime(object):
    """DateTime operations"""
    def __init__(self, expression):
        self.expression = expression


class StringOperations(object):
    """String operations"""
    def __init__(self, expression):
        self.expression = expression


class StringOperationsPandas(object):
    """String operations using Pandas Series"""
    def __init__(self, expression):
        self.expression = expression


class Expression(with_metaclass(Meta)):
    """Expression class"""
    def __init__(self, ds, expression):
        self.ds = ds
        assert not isinstance(ds, Expression)
        if isinstance(expression, Expression):
            expression = expression.expression
        self.expression = expression

    @property
    def dt(self):
        return DateTime(self)

    @property
    def str(self):
        """Gives access to string operations"""
        return StringOperations(self)

    @property
    def str_pandas(self):
        """Gives access to string operations (using Pandas Series)"""
        return StringOperationsPandas(self)

    @property
    def values(self):
        return self.evaluate()

    @property
    def dtype(self):
        return self.ds.dtype(self.expression)

    def derivative(self, var, simplify=True):
        var = _ensure_string_from_expression(var)
        return self.__class__(self.ds, expresso.derivative(self.expression, var, simplify=simplify))

    def expand(self, stop=[]):
        """Expand the expression such that no virtual columns occurs, only normal columns.

        Example:

        >>> df = vaex.example()
        >>> r = np.sqrt(df.data.x**2 + df.data.y**2)
        >>> r.expand().expression
        'sqrt(((x ** 2) + (y ** 2)))'

        """
        stop = _ensure_strings_from_expressions(stop)
        def translate(id):
            if id in self.ds.virtual_columns and id not in stop:
                return self.ds.virtual_columns[id]
        expr = expresso.translate(self.expression, translate)
        return Expression(self.ds, expr)

    def variables(self):
        """Return a set of variables this expression depends on.

        Example:

        >>> df = vaex.example()
        >>> r = np.sqrt(df.data.x**2 + df.data.y**2)
        >>> r.variables()
        {'x', 'y'}
        """
        variables = set()
        def record(varname):
            # do this recursively for virtual columns
            if varname in self.ds.virtual_columns and varname not in variables:
                expresso.translate(self.ds.virtual_columns[varname], record)
            # we don't want to record ourself
            if varname != self.expression:
                variables.add(varname)

        expresso.translate(self.expression, record)
        return variables

    def _graph(self):
        """"Return a graph containing the dependencies of this expression
        Structure is:
            [<string expression>, <function name if callable>, <function object if callable>, [subgraph/dependencies, ....]]
        """
        expression = self.expression

        def walk(node):
            if isinstance(node, six.string_types):
                if node in self.ds.virtual_columns:
                    ex = Expression(self.ds, self.ds.virtual_columns[node])
                    return [node, None, None, [ex._graph()]]
                else:
                    return node
            else:
                fname, node_repr, deps = node
                if len(node_repr) > 30:  # clip too long expressions
                    node_repr = node_repr[:26] + ' ....'
                deps = [walk(dep) for dep in deps]
                obj = self.ds.functions.get(fname)
                # we don't want the wrapper, we want the underlying object
                if isinstance(obj, Function):
                    obj = obj.f
                if isinstance(obj, FunctionSerializablePickle):
                    obj = obj.f
                return [node_repr, fname, obj, deps]
        return walk(expresso._graph(expression))

    def _graphviz(self, dot=None):
        """Return a graphviz.Digraph object with a graph of the expression"""
        from graphviz import Graph, Digraph
        node = self._graph()
        dot = dot or Digraph(comment=self.expression)
        def walk(node):
            if isinstance(node, six.string_types):
                dot.node(node, node)
                return node, node
            else:
                node_repr, fname, fobj, deps = node
                node_id = node_repr
                dot.node(node_id, node_repr)
                for dep in deps:
                    dep_id, dep = walk(dep)
                    dot.edge(node_id, dep_id)
                return node_id, node
        walk(node)
        return dot

    def __str__(self):
        return self.expression

    # def __array__(self, dtype=None):
    #     '''For casting to a numpy array

    #     Example:
    #         >>> np.array(ds.x**2)

    #     '''
    #     return self.ds.evaluate(self)

    def tolist(self):
        '''Short for expr.evaluate().tolist()'''
        return self.evaluate().tolist()

    def __repr__(self):
        return self._repr_plain_()

    def _repr_plain_(self):
        from .formatting import _format_value
        def format(values):
            for i in range(len(values)):
                value = values[i]
                yield _format_value(value)
        colalign = ("right",) * 2
        try:
            N = len(self.ds)
            if N <= PRINT_MAX_COUNT:
                values = format(self.evaluate(0, N))
                values = tabulate.tabulate([[i, k] for i, k in enumerate(values)], tablefmt='plain', colalign=colalign)
            else:
                values_head = format(self.evaluate(0, PRINT_MAX_COUNT//2))
                values_tail = format(self.evaluate(N - PRINT_MAX_COUNT//2, N))
                values_head = list(zip(range(PRINT_MAX_COUNT//2), values_head)) +\
                              list(zip(range(N - PRINT_MAX_COUNT//2, N), values_tail))
                values = tabulate.tabulate([k for k in values_head], tablefmt='plain', colalign=colalign)
                values = values.split('\n')
                width = max(map(len, values))
                separator = '\n' + '...'.center(width, ' ') + '\n'
                values = "\n".join(values[:PRINT_MAX_COUNT//2]) + separator + "\n".join(values[PRINT_MAX_COUNT//2:]) + '\n'
        except Exception as e:
            values = 'Error evaluating: %r' % e
        expression = self.expression
        if len(expression) > 60:
            expression = expression[:57] + '...'
        info = 'Expression = ' + expression + '\n'
        str_type = str
        dtype = self.dtype
        dtype = (str(dtype) if dtype != str_type else 'str')
        if self.expression in self.ds.get_column_names(hidden=True):
            state = "column"
        elif self.expression in self.ds.get_column_names(hidden=True):
            state = "virtual column"
        else:
            state = "expression"
        line = 'Length: {:,} dtype: {} ({})\n'.format(len(self.ds), dtype, state)
        info += line
        info += '-' * (len(line)-1) + '\n'
        info += values
        return info

    def count(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, edges=False, progress=None):
        '''Shortcut for ds.count(expression, ...), see `Dataset.count`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.count(**kwargs)

    def sum(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.sum(expression, ...), see `Dataset.sum`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.sum(**kwargs)

    def mean(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.mean(expression, ...), see `Dataset.mean`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.mean(**kwargs)

    def std(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.std(expression, ...), see `Dataset.std`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.std(**kwargs)

    def var(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.std(expression, ...), see `Dataset.var`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.var(**kwargs)

    def minmax(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.minmax(expression, ...), see `Dataset.minmax`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.minmax(**kwargs)

    def min(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.min(expression, ...), see `Dataset.min`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.min(**kwargs)

    def max(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.max(expression, ...), see `Dataset.max`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.max(**kwargs)

    def nop(self):
        """Evaluates expression, and drop the result, usefull for benchmarking, since vaex is usually lazy"""
        return self.ds.nop(self.expression)

    @property
    def transient(self):
        """If this expression is not transient (e.g. on disk) optimizations can be made"""
        return self.expand().expression not in self.ds.columns

    @property
    def masked(self):
        """Alias to df.is_masked(expression)"""
        return self.ds.is_masked(self.expression)

    def value_counts(self, dropna=False, dropnull=True, ascending=False, progress=False):
        """Computes counts of unique values.

         WARNING:
          * If the expression/column is not categorical, it will be converted on the fly
          * dropna is False by default, it is True by default in pandas

        :param dropna: when True, it will not report the missing values
        :param ascending: when False (default) it will report the most frequent occuring item first
        :returns: Pandas series containing the counts
        """
        from pandas import Series
        dtype = self.dtype

        transient = self.transient or self.ds.filtered or self.ds.is_masked(self.expression)
        if self.dtype == str_type and not transient:
            # string is a special case, only ColumnString are not transient
            ar = self.ds.columns[self.expression]
            if not isinstance(ar, ColumnString):
                transient = True

        counter_type = counter_type_from_dtype(self.dtype, transient)
        counters = [None] * self.ds.executor.thread_pool.nthreads
        def map(thread_index, i1, i2, ar):
            if counters[thread_index] is None:
                counters[thread_index] = counter_type()
            if dtype == str_type:
                previous_ar = ar
                ar = _to_string_sequence(ar)
                if not transient:
                    assert ar is previous_ar.string_sequence
            if np.ma.isMaskedArray(ar):
                mask = np.ma.getmaskarray(ar)
                counters[thread_index].update(ar, mask)
            else:
                counters[thread_index].update(ar)
            return 0
        def reduce(a, b):
            return a+b
        self.ds.map_reduce(map, reduce, [self.expression], delay=False, progress=progress, name='value_counts', info=True, to_numpy=False)
        counters = [k for k in counters if k is not None]
        counter0 = counters[0]
        for other in counters[1:]:
            counter0.merge(other)
        value_counts = counter0.extract()
        index = np.array(list(value_counts.keys()))
        counts = np.array(list(value_counts.values()))

        order = np.argsort(counts)
        if not ascending:
            order = order[::-1]
        counts = counts[order]
        index = index[order]
        if not dropna or not dropnull:
            index = index.tolist()
            counts = counts.tolist()
            if not dropna and counter0.nan_count:
                index = [np.nan] + index
                counts = [counter0.nan_count] + counts
            if not dropnull and counter0.null_count:
                index = ['null'] + index
                counts = [counter0.null_count] + counts

        return Series(counts, index=index)

    def unique(self):
        return self.ds.unique(self.expression)

    def evaluate(self, i1=None, i2=None, out=None, selection=None):
        return self.ds.evaluate(self, i1, i2, out=out, selection=selection)

    # TODO: it is not so elegant we need to have a custom version of this
    # it now also misses the docstring, reconsider how the the meta class auto
    # adds this method
    def fillna(self, value, fill_nan=True, fill_masked=True):
        return self.ds.func.fillna(self, value=value, fill_nan=fill_nan, fill_masked=fill_masked)

    def clip(self, lower=None, upper=None):
        return self.ds.func.clip(self, lower, upper)

    def jit_numba(self, verbose=False):
        import imp
        import hashlib
        names = []
        funcs = set(expression_namespace.keys())
        # if it's a virtual column, we probably want to optimize that
        # TODO: fully extract the virtual columns, i.e. depending ones?
        expression = self.expression
        if expression in self.ds.virtual_columns:
            expression = self.ds.virtual_columns[self.expression]
        all_vars = self.ds.get_column_names(virtual=True, strings=True, hidden=True) + list(self.ds.variables.keys())
        vaex.expresso.validate_expression(expression, all_vars, funcs, names)
        arguments = list(set(names))
        argument_dtypes = [self.ds.dtype(argument) for argument in arguments]
        # argument_dtypes = [getattr(np, dtype_name) for dtype_name in dtype_names]

        # TODO: for now only float64 output supported
        f = FunctionSerializableNumba(expression, arguments, argument_dtypes, return_dtype=np.dtype(np.float64))
        function = self.ds.add_function('_jit', f, unique=True)
        return function(*arguments)

    def jit_pythran(self, verbose=False):
        import logging
        logger = logging.getLogger('pythran')
        log_level = logger.getEffectiveLevel()
        try:
            if not verbose:
                logger.setLevel(logging.ERROR)
            import pythran
            import imp
            import hashlib
            # self._import_all(module)
            names = []
            funcs = set(expression_namespace.keys())
            expression = self.expression
            if expression in self.ds.virtual_columns:
                expression = self.ds.virtual_columns[self.expression]
            all_vars = self.ds.get_column_names(virtual=True, strings=True, hidden=True) + list(self.ds.variables.keys())
            vaex.expresso.validate_expression(expression, all_vars, funcs, names)
            names = list(set(names))
            types = ", ".join(str(self.ds.dtype(name)) + "[]" for name in names)
            argstring = ", ".join(names)
            code = '''
from numpy import *
#pythran export f({2})
def f({0}):
    return {1}'''.format(argstring, expression, types)
            if verbose:
                print("generated code")
                print(code)
            m = hashlib.md5()
            m.update(code.encode('utf-8'))
            module_name = "pythranized_" + m.hexdigest()
            # print(m.hexdigest())
            module_path = pythran.compile_pythrancode(module_name, code, extra_compile_args=["-DBOOST_SIMD", "-march=native"] + [] if verbose else ["-w"])

            module = imp.load_dynamic(module_name, module_path)
            function_name = "f_" + m.hexdigest()
            expression_namespace[function_name] = module.f

            return Expression(self.ds, "{0}({1})".format(function_name, argstring))
        finally:
                logger.setLevel(log_level)

    def _rename(self, old, new):
        def translate(id):
            if id == old:
                return new
        expr = expresso.translate(self.expression, translate)
        return Expression(self.ds, expr)

    def astype(self, dtype):
        return self.ds.func.astype(self, str(dtype))

    def apply(self, f):
        return self.ds.apply(f, [self.expression])

    def map(self, mapper, nan_value=None, null_value=None, default_value=None, allow_missing=False):
        """Map values of an expression or in memory column accoring to an input
        dictionary or a custom callable function.

        Example:

        >>> import vaex
        >>> df = vaex.from_arrays(color=['red', 'red', 'blue', 'red', 'green'])
        >>> mapper = {'red': 1, 'blue': 2, 'green': 3}
        >>> df['color_mapped'] = df.color.map(mapper)
        >>> df
        #  color      color_mapped
        0  red                   1
        1  red                   1
        2  blue                  2
        3  red                   1
        4  green                 3
        >>> import numpy as np
        >>> df = vaex.from_arrays(type=[0, 1, 2, 2, 2, np.nan])
        >>> df['role'] = df['type'].map({0: 'admin', 1: 'maintainer', 2: 'user', np.nan: 'unknown'})
        >>> df
        #    type  role
        0       0  admin
        1       1  maintainer
        2       2  user
        3       2  user
        4       2  user
        5     nan  unknown
        >>> import vaex
        >>> import numpy as np
        >>> df = vaex.from_arrays(type=[0, 1, 2, 2, 2, 4])
        >>> df['role'] = df['type'].map({0: 'admin', 1: 'maintainer', 2: 'user'}, default_value='unknown')
        >>> df
        #    type  role
        0       0  admin
        1       1  maintainer
        2       2  user
        3       2  user
        4       2  user
        5       4  unknown
        :param mapper: dict like object used to map the values from keys to values
        :param nan_value: value to be used when a nan is present (and not in the mapper)
        :param null_value: value to use used when there is a missing value
        :param default_value: value to be used when a value is not in the mapper (like dict.get(key, default))
        :param allow_missing: used to signal that values in the mapper should map to a masked array with missing values,
            assumed True when default_value is not None.
        :return: A vaex expression
        :rtype: vaex.expression.Expression
        """
        assert isinstance(mapper, collectionsAbc.Mapping), "mapper should be a dict like object"

        df = self.ds
        mapper_keys = list(mapper.keys())
        try:
            mapper_nan_key_mask = np.isnan(mapper_keys)
        except TypeError:
            # case where we have mixed strings/nan etc
            def try_nan(x):
                try:
                    return np.isnan(x)
                except:
                    return False
            mapper_nan_key_mask = np.array([try_nan(k) for k in mapper_keys])
        mapper_has_nan = mapper_nan_key_mask.sum() > 0
        if mapper_nan_key_mask.sum() > 1:
            raise ValueError('Insanity, you provided multiple nan values as keys for your dict')

        # we map the keys to a ordinal values [0, N-1] using the set
        key_set = df._set(self.expression)
        found_keys = key_set.keys()

        # we want all possible values to be converted
        # so mapper's key should be a superset of the keys found
        use_masked_array = False
        if default_value is not None:
            allow_missing = True
        if not set(mapper_keys).issuperset(found_keys):
            missing = set(found_keys).difference(mapper_keys)
            missing0 = list(missing)[0]
            only_has_nan = missing0 != missing0 and len(missing) == 1
            if allow_missing:
                if default_value is not None:
                    value0 = list(mapper.values())[0]
                    assert np.issubdtype(type(default_value), np.array(value0).dtype), "default value has to be of similar type"
                else:
                    use_masked_array = True
            else:
                if only_has_nan:
                    pass  # we're good, the hash mapper deals with nan
                else:
                    raise ValueError('Missing values in mapper: %s' % missing)

        # and these are the corresponding choices
        # note that here we map 'planned' unknown values to the default values
        # and later on in _choose, we map values not even seen in the dataframe
        # to the default_value
        choices = [mapper.get(key, default_value) for key in found_keys]
        if key_set.has_nan:
            if mapper_has_nan:
                # since np.nan is not np.nan/2, we have to use the real key object
                nan_index = np.arange(len(mapper_keys))[mapper_nan_key_mask][0]
                nan_key_used = mapper_keys[nan_index]
                choices = [mapper[nan_key_used]] + choices
            else:
                choices = [nan_value] + choices
        if key_set.has_null:
            choices = [null_value] + choices
        choices = np.array(choices)

        key_set_name = df.add_variable('map_key_set', key_set, unique=True)
        choices_name = df.add_variable('map_choices', choices, unique=True)
        if allow_missing:
            if use_masked_array:
                expr = '_choose_masked(_ordinal_values({}, {}), {})'.format(self, key_set_name, choices_name)
            else:
                expr = '_choose(_ordinal_values({}, {}), {}, {!r})'.format(self, key_set_name, choices_name, default_value)
        else:
            expr = '_choose(_ordinal_values({}, {}), {})'.format(self, key_set_name, choices_name)
        return Expression(df, expr)


class FunctionSerializable(object):
    pass


@vaex.serialize.register
class FunctionSerializablePickle(FunctionSerializable):
    def __init__(self, f=None):
        self.f = f

    def pickle(self, function):
        return pickle.dumps(function)

    def unpickle(self, data):
        return pickle.loads(data)

    def state_get(self):
        data = self.pickle(self.f)
        if vaex.utils.PY2:
            pickled = base64.encodestring(data)
        else:
            pickled = base64.encodebytes(data).decode('ascii')
        return dict(pickled=pickled)

    @classmethod
    def state_from(cls, state):
        obj = cls()
        obj.state_set(state)
        return obj

    def state_set(self, state):
        data = state['pickled']
        if vaex.utils.PY2:
            data = base64.decodestring(data)
        else:
            data = base64.decodebytes(data.encode('ascii'))
        self.f = self.unpickle(data)

    def __call__(self, *args, **kwargs):
        '''Forward the call to the real function'''
        return self.f(*args, **kwargs)


@vaex.serialize.register
class FunctionSerializableNumba(FunctionSerializable):
    def __init__(self, expression, arguments, argument_dtypes, return_dtype, verbose=False):
        self.expression = expression
        self.arguments = arguments
        self.argument_dtypes = argument_dtypes
        self.return_dtype = return_dtype
        self.verbose = verbose
        import numba
        argument_dtypes_numba = [getattr(numba, argument_dtype.name) for argument_dtype in argument_dtypes]
        argstring = ", ".join(arguments)
        code = '''
from numpy import *
def f({0}):
    return {1}'''.format(argstring, expression)
        if verbose:
            print('Generated code:\n' + code)
        scope = {}
        exec(code, scope)
        f = scope['f']
        return_dtype_numba = getattr(numba, return_dtype.name)
        vectorizer = numba.vectorize([return_dtype_numba(*argument_dtypes_numba)])
        self.f = vectorizer(f)

    def __call__(self, *args, **kwargs):
        '''Forward the call to the numba function'''
        return self.f(*args, **kwargs)

    def state_get(self):
        return dict(expression=self.expression,
                    arguments=self.arguments,
                    argument_dtypes=list(map(str, self.argument_dtypes)),
                    return_dtype=str(self.return_dtype),
                    verbose=self.verbose)

    @classmethod
    def state_from(cls, state):
        return cls(expression=state['expression'],
                   arguments=state['arguments'],
                   argument_dtypes=list(map(np.dtype, state['argument_dtypes'])),
                   return_dtype=np.dtype(state['return_dtype']),
                   verbose=state['verbose'])


# TODO: this is not the right abstraction, since this won't allow a
# numba version for the function
@vaex.serialize.register
class FunctionToScalar(FunctionSerializablePickle):
    def __call__(self, *args, **kwargs):
        length = len(args[0])
        result = []
        def fix_type(v):
            # TODO: only when column is str type?
            if isinstance(v, np.str_):
                return str(v)
            if isinstance(v, np.bytes_):
                return v.decode('utf8')
            else:
                return v
        for i in range(length):
            scalar_result = self.f(*[fix_type(k[i]) for k in args], **{key: value[i] for key, value in kwargs.items()})
            result.append(scalar_result)
        result = np.array(result)
        return result


class Function(object):

    def __init__(self, dataset, name, f):
        self.dataset = dataset
        self.name = name

        if not vaex.serialize.can_serialize(f): # if not serializable, assume we can use pickle
            f = FunctionSerializablePickle(f)
        self.f = f

    def __call__(self, *args, **kwargs):
        arg_string = ", ".join([str(k) for k in args] + ['{}={:r}'.format(name, value) for name, value in kwargs.items()])
        expression = "{}({})".format(self.name, arg_string)
        return Expression(self.dataset, expression)


class FunctionBuiltin(object):

    def __init__(self, dataset, name, **kwargs):
        self.dataset = dataset
        self.name = name
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        kwargs = dict(kwargs, **self.kwargs)
        arg_string = ", ".join([str(k) for k in args] + ['{}={:r}'.format(name, value) for name, value in kwargs.items()])
        expression = "{}({})".format(self.name, arg_string)
        return Expression(self.dataset, expression)



from __future__ import absolute_import
__author__ = 'breddels'
import numpy as np
import logging
import threading
import uuid
import time
import ast
from .dataset import Dataset
from .utils import _issequence
from .tasks import Task
from .legacy import Subspace
import vaex.promise
import vaex.settings
import vaex.utils
from tornado.httpclient import AsyncHTTPClient, HTTPClient
import tornado.httputil
import tornado.websocket
from tornado.concurrent import Future
from tornado import gen
from .dataset import default_shape
import tornado.ioloop
import json
import astropy.units
from vaex.utils import _ensure_strings_from_expressions, _ensure_string_from_expression


try:
    import __builtin__
except ImportError:
    import builtins as __builtin__


try:
    import Cookie  # py2
except ImportError:
    import http.cookies as Cookie  # py3


logger = logging.getLogger("vaex.remote")

DEFAULT_REQUEST_TIMEOUT = 60 * 5  # 5 minutes


def wrap_future_with_promise(future):
    if isinstance(future, vaex.promise.Promise):  # TODO: not so nice, sometimes we pass a promise
        return future
    promise = vaex.promise.Promise()

    def callback(future):
        # print(("callback", future, future.result()))
        e = future.exception()
        if e:
            # print(("reject", e))
            promise.reject(e)
        else:
            promise.fulfill(future.result())
    future.add_done_callback(callback)
    return promise


def listify(value):
    # TODO: listify is a bad name, can we use a common serialization function?
    if isinstance(value, vaex.expression.Expression):
        return str(value)
    elif isinstance(value, list):
        value = list([listify(item) for item in value])
    elif hasattr(value, "tolist"):
        value = value.tolist()
    return value


try:
    from urllib.request import urlopen
    from urllib.parse import urlparse, urlencode
except ImportError:
    from urlparse import urlparse
    from urllib import urlopen, urlencode


def _check_error(object):
    if "error" in object:
        raise RuntimeError("Server responded with error: %r" % object["error"])


class ServerRest(object):
    def __init__(self, hostname, port=5000, base_path="/", background=False, thread_mover=None, websocket=True):
        self.hostname = hostname
        self.port = port
        self.base_path = base_path if base_path.endswith("/") else (base_path + "/")
        # if delay:
        event = threading.Event()
        self.thread_mover = thread_mover or (lambda fn, *args, **kwargs: fn(*args, **kwargs))
        logger.debug("thread mover: %r", self.thread_mover)

        # jobs maps from uid to tasks
        self.jobs = {}

        def ioloop_threaded():
            logger.debug("creating tornado io_loop")
            self.io_loop = tornado.ioloop.IOLoop().current()
            event.set()
            logger.debug("started tornado io_loop...")

            self.io_loop.start()
            self.io_loop.close()
            logger.debug("stopped tornado io_loop")

        io_loop = tornado.ioloop.IOLoop.current(instance=False)
        if True:  # io_loop:# is None:
            logger.debug("no current io loop, starting it in thread")
            thread = threading.Thread(target=ioloop_threaded)
            thread.setDaemon(True)
            thread.start()
            event.wait()
        else:
            logger.debug("using current io loop")
            self.io_loop = io_loop

        self.io_loop.make_current()
        # if async:
        self.http_client_async = AsyncHTTPClient()
        # self.http_client = HTTPClient()
        self.user_id = vaex.settings.webclient.get("cookie.user_id")
        self.use_websocket = websocket
        self.websocket = None

        self.submit = self.submit_http

        if self.use_websocket:
            logger.debug("connect via websocket")
            self.submit = self.submit_websocket
            self._websocket_connect()
            logger.debug("websocket connected")

    def close(self):
        # self.http_client_async.
        # self.http_client.close()
        if self.use_websocket:
            if self.websocket:
                self.websocket.close()
        self.io_loop.stop()

    def _websocket_connect(self):
        def connected(websocket):
            logger.debug("connected to websocket: %s" % self._build_url(""))

        def failed(reason):
            logger.error("failed to connect to %s" % self._build_url(""))
        self.websocket_connected = vaex.promise.Promise()
        self.websocket_connected.then(connected, failed)
        if 0:
            connected = wrap_future_with_promise(tornado.websocket.websocket_connect(self._build_url("websocket"), on_message_callback=self._on_websocket_message))
            connected.get()
            self.websocket_connected.fulfill(connected)

        def do():
            try:
                logger.debug("wrapping promise")
                connected = wrap_future_with_promise(tornado.websocket.websocket_connect(self._build_url("websocket"), on_message_callback=self._on_websocket_message))
                logger.debug("continue")
                self.websocket_connected.fulfill(connected)
            except:
                logger.exception("error connecting")
                # raise
        logger.debug("add callback")
        self.io_loop.add_callback(do)
        logger.debug("added callback: ")
        # self.io_loop.start()
        # if self.port == 29345:
        # import pdb
        # pdb.set_trace()
        logger.debug("waiting for connection")
        result = self.websocket_connected.get()
        logger.debug("websocket connected")
        if self.websocket_connected.isRejected:
            raise self.websocket.reason

    def _on_websocket_message(self, msg):
        try:
            task = None
            json_data, data = msg.split(b"\n", 1)
            response = json.loads(json_data.decode("utf8"))
            phase = response["job_phase"]
            job_id = response.get("job_id")
            task = self.jobs[job_id]
            if data:
                import zlib
                data = zlib.decompress(data)
                try:
                    numpy_array = np.frombuffer(data, dtype=np.dtype(response["dtype"])).reshape(ast.literal_eval(response["shape"]))
                except:
                    logger.exception("error in decoding data: %r %r %r", data, response, task.task_queue)
                finally:
                    response["result"] = numpy_array
            import sys
            if sys.getsizeof(msg) > 1024 * 4:
                logger.debug("socket read message: <large amount of data>",)
            else:
                logger.debug("socket read message: %s", msg)
                logger.debug("json response: %r", response)
            # for the moment, job == task, in the future a job can be multiple tasks
        except Exception as e:
            if task:
                task.reject(e)
            logger.exception("unexpected decoding error")
            return
        if job_id:
            try:
                logger.debug("job update %r, phase=%r", job_id, phase)
                if phase == "COMPLETED":
                    result = response["result"]  # [0]
                    # logger.debug("completed job %r, result=%r", job_id, result)
                    logger.debug("completed job %r (delay=%r, thread_mover=%r)", job_id, task.delay, self.thread_mover)
                    processed_result = task.post_process(result)
                    if task.delay:
                        self.thread_mover(task.fulfill, processed_result)
                    else:
                        task.fulfill(processed_result)
                elif phase == "EXCEPTION":
                    logger.error("exception happened at server side: %r", response)
                    class_name = response["exception"]["class"]
                    msg = response["exception"]["msg"]
                    exception = getattr(__builtin__, class_name)(msg)
                    logger.debug("error in job %r, exception=%r", job_id, exception)
                    task = self.jobs[job_id]
                    if task.delay:
                        self.thread_mover(task.reject, exception)
                    else:
                        task.reject(exception)
                elif phase == "ERROR":
                    logger.error("error happened at server side: %r", response)
                    msg = response["error"]
                    exception = RuntimeError("error at server: %r" % msg)
                    task = self.jobs[job_id]
                    if task.delay:
                        self.thread_mover(task.reject, exception)
                    else:
                        task.reject(exception)
                elif phase == "PENDING":
                    fraction = response["progress"]
                    logger.debug("pending?: %r", phase)
                    task = self.jobs[job_id]
                    if task.delay:
                        self.thread_mover(task.signal_progress.emit, fraction)
                    else:
                        task.signal_progress.emit(fraction)
            except Exception as e:
                logger.exception("error in handling job")
                task = self.jobs[job_id]
                if task.delay:
                    self.thread_mover(task.reject, e)
                else:
                    task.reject(e)

    def wait(self):
        io_loop = tornado.ioloop.IOLoop.instance()
        io_loop.start()

    def submit_websocket(self, path, arguments, delay=False, progress=None, post_process=lambda x: x):
        assert self.use_websocket

        task = TaskServer(post_process=post_process, delay=delay)
        progressbars = vaex.utils.progressbars(progress)
        progressbars.add_task(task)
        logger.debug("created task: %r, %r (delay=%r)" % (path, arguments, delay))
        job_id = str(uuid.uuid4())
        self.jobs[job_id] = task
        arguments["job_id"] = job_id
        arguments["path"] = path
        arguments["user_id"] = self.user_id
        # arguments = dict({key: (value.tolist() if hasattr(value, "tolist") else value) for key, value in arguments.items()})
        arguments = dict({key: listify(value) for key, value in arguments.items()})

        def do():
            def write(socket):
                try:
                    logger.debug("write to websocket: %r", arguments)
                    socket.write_message(json.dumps(arguments))
                    return
                except:
                    import traceback
                    traceback.print_exc()
            # return
            logger.debug("will schedule a write to the websocket")
            self.websocket_connected.then(write).end()  # .then(task.fulfill)

        self.io_loop.add_callback(do)
        logger.debug("we can continue (main thread is %r)", threading.currentThread())
        if delay:
            return task
        else:
            return task.get()

    def submit_http(self, path, arguments, post_process, delay, progress=None, **kwargs):
        def pre_post_process(response):
            cookie = Cookie.SimpleCookie()
            for cookieset in response.headers.get_list("Set-Cookie"):
                cookie.load(cookieset)
                logger.debug("cookie load: %r", cookieset)
            logger.debug("cookie: %r", cookie)
            if "user_id" in cookie:
                user_id = cookie["user_id"].value
                logger.debug("user_id: %s", user_id)
                if self.user_id != user_id:
                    self.user_id = user_id
                    vaex.settings.webclient.store("cookie.user_id", self.user_id)
            data = response.body
            is_json = False
            logger.info("response is: %r", response.body)
            logger.info("content_type is: %r", response.headers["Content-Type"])
            if response.headers["Content-Type"] == "application/numpy-array":
                shape, dtype, data = response.body.split(b"\n", 2)
                shape = shape.decode("ascii")
                dtype = dtype.decode("ascii")
                import ast
                numpy_array = np.fromstring(data, dtype=np.dtype(dtype)).reshape(ast.literal_eval(shape))
                return post_process(numpy_array)
            else:
                try:
                    data = json.loads(response.body.decode("ascii"))
                    is_json = True
                except Exception as e:
                    logger.info("couldn't convert to json (error is %s, assume it's raw data): %s", e, data)
                    # logger.info("couldn't convert to json (error is %s, assume it's raw data)", e)
                if is_json:
                    self._check_exception(data)
                    return post_process(data["result"])
                else:
                    return post_process(data)

        arguments = {key: listify(value) for key, value in arguments.items()}
        import pdb
        arguments_json = {key: json.dumps(value) for key, value in arguments.items()}
        headers = tornado.httputil.HTTPHeaders()

        url = self._build_url(path + "?" + urlencode(arguments_json))
        logger.debug("fetch %s, delay=%r", url, delay)

        if self.user_id is not None:
            headers.add("Cookie", "user_id=%s" % self.user_id)
            logger.debug("adding user_id %s to request", self.user_id)
        if delay:
            task = TaskServer(pre_post_process, delay=delay)
            # tornado doesn't like that we call fetch while ioloop is running in another thread, we should use ioloop.add_callbacl

            def do():
                self.thread_mover(task.signal_progress.emit, 0.5)
                future = self.http_client_async.fetch(url, headers=headers, request_timeout=DEFAULT_REQUEST_TIMEOUT, **kwargs)

                def thread_save_succes(value):
                    self.thread_mover(task.signal_progress.emit, 1.0)
                    self.thread_mover(task.fulfill, value)

                def thread_save_failure(value):
                    self.thread_mover(task.reject, value)
                wrap_future_with_promise(future).then(pre_post_process).then(thread_save_succes, thread_save_failure)
            self.io_loop.add_callback(do)
            return task  # promise.then(self._move_to_thread)
        else:
            return pre_post_process(self.http_client.fetch(url, headers=headers, request_timeout=DEFAULT_REQUEST_TIMEOUT, **kwargs))

    def datasets(self, as_dict=False, delay=False):
        def post(result):
            logger.debug("datasets result: %r", result)

            def create(server, state):
                dataset = DatasetRest(self, name=state['name'],
                                      length_original=state['length_original'],
                                      column_names=state['column_names'],
                                      dtypes=state['dtypes'])
                dataset.state_set(state['state'])
                return dataset
            datasets = [create(self, kwargs) for kwargs in result]
            logger.debug("datasets: %r", datasets)
            return datasets if not as_dict else dict([(ds.name, ds) for ds in datasets])
        return self.submit(path="datasets", arguments={}, post_process=post, delay=delay)

    def _build_url(self, method):
        protocol = "ws" if self.use_websocket else "http"
        return "%s://%s:%d%s%s" % (protocol, self.hostname, self.port, self.base_path, method)

    def _call_subspace(self, method_name, subspace, **kwargs):
        def post_process(result):
            if method_name == "histogram":  # histogram is the exception..
                # TODO: don't do binary transfer, just json, now we cannot handle exception
                import base64
                # logger.debug("result: %r", result)
                # result = base64.b64decode(result) #.decode("base64")
                # result = base64.
                data = np.fromstring(result, dtype=np.float64)
                shape = (kwargs["size"],) * subspace.dimension
                data = data.reshape(shape)
                return data
            else:
                try:
                    return np.array(result)
                except ValueError:
                    return result
        dataset_name = subspace.dataset.name
        expressions = subspace.expressions
        delay = subspace.delay
        path = "datasets/%s/%s" % (dataset_name, method_name)
        url = self._build_url(path)
        arguments = dict(kwargs)
        dataset_remote = subspace.dataset
        arguments["selection"] = subspace.is_masked
        arguments['state'] = dataset_remote.state_get()
        arguments['auto_fraction'] = dataset_remote.get_auto_fraction()
        arguments.update(dict(expressions=expressions))
        return self.submit(path, arguments, post_process=post_process, delay=delay)

    def _call_dataset(self, method_name, dataset_remote, delay, numpy=False, progress=None, **kwargs):
        def post_process(result):
            # result = self._check_exception(json.loads(result.body))["result"]
            if numpy:
                return np.fromstring(result, dtype=np.float64)
            else:
                try:
                    return np.array(result)
                except ValueError:
                    return result
        path = "datasets/%s/%s" % (dataset_remote.name, method_name)
        arguments = dict(kwargs)
        arguments['state'] = dataset_remote.state_get()
        arguments['auto_fraction'] = dataset_remote.get_auto_fraction()
        body = urlencode(arguments)

        return self.submit(path, arguments, post_process=post_process, progress=progress, delay=delay)

    def _schedule_call(self, method_name, dataset_remote, delay, **kwargs):
        def post_process(result):
            # result = self._check_exception(json.loads(result.body))["result"]
            try:
                return np.array(result)
            except ValueError:
                return result
        method = "%s/%s" % (dataset_remote.name, method_name)
        return self.schedule(path, arguments, post_process=post_process, delay=delay)

    def _check_exception(self, reply_json):
        if "exception" in reply_json:
            logger.error("exception happened at server side: %r", reply_json)
            class_name = reply_json["exception"]["class"]
            msg = reply_json["exception"]["msg"]
            raise getattr(__builtin__, class_name)(msg)
        if "error" in reply_json:
            raise ValueError("unknown error occured at server")
        else:
            return reply_json


class SubspaceRemote(Subspace):
    def toarray(self, list):
        return np.array(list)

    @property
    def dimension(self):
        return len(self.expressions)

    def _task(self, promise):
        """Helper function for returning tasks results, result when immediate is True, otherwise the task itself, which is a promise"""
        if self.delay:
            return promise
        else:
            return promise

    def sleep(self, seconds, delay=False):
        return self.dataset.server.call("sleep", seconds, delay=delay)

    def minmax(self):
        return self._task(self.dataset.server._call_subspace("minmax", self))
        # return self._task(task)

    def histogram(self, limits, size=256, weight=None):
        return self._task(self.dataset.server._call_subspace("histogram", self, size=size, limits=limits, weight=weight))

    def nearest(self, point, metric=None):
        point = vaex.utils.make_list(point)
        result = self.dataset.server._call_subspace("nearest", self, point=point, metric=metric)
        return self._task(result)

    def mean(self):
        return self.dataset.server._call_subspace("mean", self)

    def correlation(self, means=None, vars=None):
        return self.dataset.server._call_subspace("correlation", self, means=means, vars=vars)

    def var(self, means=None):
        return self.dataset.server._call_subspace("var", self, means=means)

    def sum(self):
        return self.dataset.server._call_subspace("sum", self)

    def limits_sigma(self, sigmas=3, square=False):
        return self.dataset.server._call_subspace("limits_sigma", self, sigmas=sigmas, square=square)

    def mutual_information(self, limits=None, size=256):
        return self.dataset.server._call_subspace("mutual_information", self, limits=limits, size=size)


class DatasetRemote(Dataset):
    def __init__(self, name, server, column_names):
        super(DatasetRemote, self).__init__(name, column_names)
        self.server = server


class DatasetRest(DatasetRemote):
    def __init__(self, server, name, column_names, dtypes, length_original):
        DatasetRemote.__init__(self, name, server.hostname, column_names)
        self.server = server
        self.name = name
        self.column_names = column_names
        self._dtypes = {name: np.dtype(dtype) for name, dtype in dtypes.items()}
        for column_name in self.get_column_names(virtual=True, strings=True):
            self._save_assign_expression(column_name)
        self._length_original = length_original
        self._length_unfiltered = length_original
        self._index_end = length_original
        self.path = self.filename = self.server._build_url("%s" % name)
        self.fraction = 1
        self.executor = ServerExecutor()

    def copy(self, column_names=None, virtual=True):
        dtypes = {name: self.dtype(name) for name in self.get_column_names(strings=True, virtual=False)}
        ds = DatasetRest(self.server, self.name, self.column_names, dtypes=dtypes, length_original=self._length_original)
        state = self.state_get()
        if not virtual:
            state['virtual_columns'] = {}
        ds.state_set(state, use_active_range=True)
        return ds

    def trim(self, inplace=False):
        df = self if inplace else self.copy()
        # can we get away with not trimming?
        return df

    def count(self, expression=None, binby=[], limits=None, shape=default_shape, selection=False, delay=False, edges=False, progress=None):
        return self._delay(delay, self.server._call_dataset("count", self, delay=True, progress=progress, expression=expression, binby=binby, limits=limits, shape=shape, selection=selection, edges=edges))

    def mean(self, expression, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        return self._delay(delay, self.server._call_dataset("mean", self, delay=True, progress=progress, expression=expression, binby=binby, limits=limits, shape=shape, selection=selection))

    def sum(self, expression, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        return self._delay(delay, self.server._call_dataset("sum", self, delay=True, progress=progress, expression=expression, binby=binby, limits=limits, shape=shape, selection=selection))

    def var(self, expression, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        return self._delay(delay, self.server._call_dataset("var", self, delay=True, progress=progress, expression=expression, binby=binby, limits=limits, shape=shape, selection=selection))

    def minmax(self, expression, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        return self._delay(delay, self.server._call_dataset("minmax", self, delay=True, progress=progress, expression=expression, binby=binby, limits=limits, shape=shape, selection=selection))

    # def count(self, expression=None, binby=[], limits=None, shape=default_shape, selection=False, delay=False):
    def cov(self, x, y=None, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        return self._delay(delay, self.server._call_dataset("cov", self, delay=True, progress=progress, x=x, y=y, binby=binby, limits=limits, shape=shape, selection=selection))

    def correlation(self, x, y=None, binby=[], limits=None, shape=default_shape, sort=False, sort_key=np.abs, selection=False, delay=False, progress=None):
        # TODO: sort and sort_key should be done locally
        return self._delay(delay, self.server._call_dataset("correlation", self, delay=True, progress=progress, x=x, y=y, binby=binby, limits=limits, shape=shape, selection=selection))

    def _delay(self, delay, task, progressbar=False):
        if delay:
            return task
        else:
            result = task.get()
            logger.debug("result = %r", result)
            return result

    def dtype(self, expression):
        if expression in self._dtypes:
            return self._dtypes[expression]
        else:
            return np.zeros(1, dtype=np.float64).dtype

    def is_local(self): return False

    def __repr__(self):
        name = self.__class__.__module__ + "." + self.__class__.__name__
        return "<%s(server=%r, name=%r, column_names=%r, __len__=%r)> instance at 0x%x" % (name, self.server, self.name, self.column_names, len(self), id(self))

    def __call__(self, *expressions, **kwargs):
        return SubspaceRemote(self, expressions, kwargs.get("executor") or self.executor, delay=kwargs.get("delay", False))

    def evaluate(self, expression, i1=None, i2=None, out=None, selection=None, delay=False):
        expression = _ensure_strings_from_expressions(expression)
        """basic support for evaluate at server, at least to run some unittest, do not expect this to work from strings"""
        result = self.server._call_dataset("evaluate", self, expression=expression, i1=i1, i2=i2, selection=selection, delay=delay)
        # TODO: we ignore out
        return result

    def execute(self):
        '''Execute all delayed jobs.'''
        self.executor.execute()
        # TODO: should be support _task_agg? If we do, we can use the base class' method
        # self._task_aggs.clear()


# we may get rid of this when we group together tasks
class ServerExecutor(object):
    def __init__(self):
        self.signal_begin = vaex.events.Signal("begin")
        self.signal_progress = vaex.events.Signal("progress")
        self.signal_end = vaex.events.Signal("end")
        self.signal_cancel = vaex.events.Signal("cancel")

    def execute(self):
        logger.debug("dummy execute")


class TaskServer(Task):
    def __init__(self, post_process, delay):
        Task.__init__(self, None, [])
        self.post_process = post_process
        self.delay = delay
        self.task_queue = []

    def schedule(self, task):
        self.task_queue.append(task)
        logger.info("task added, queue: %r", self.task_queue)
        return task

    def execute(self):
        logger.debug("starting with execute")
        if self._is_executing:
            logger.debug("nested execute call")
            # this situation may happen since in this methods, via a callback (to update a progressbar) we enter
            # Qt's eventloop, which may execute code that will call execute again
            # as long as that code is using delay tasks (i.e. promises) we can simple return here, since after
            # the execute is almost finished, any new tasks added to the task_queue will get executing
            return
        # u 'column' is uniquely identified by a tuple of (dataset, expression)
        self._is_executing = True
        try:
            t0 = time.time()
            task_queue_all = list(self.task_queue)
            if not task_queue_all:
                logger.info("only had cancelled tasks")
            logger.info("clearing queue")
            # self.task_queue = [] # Ok, this was stupid.. in the meantime there may have been new tasks, instead, remove the ones we copied
            for task in task_queue_all:
                logger.info("remove from queue: %r", task)
                self.task_queue.remove(task)
            logger.info("left in queue: %r", self.task_queue)
            task_queue_all = [task for task in task_queue_all if not task.cancelled]
            logger.debug("executing queue: %r" % (task_queue_all))

            # for task in self.task_queue:
            # $	print task, task.expressions_all
            datasets = set(task.dataset for task in task_queue_all)
            cancelled = [False]

            def cancel():
                logger.debug("cancelling")
                self.signal_cancel.emit()
                cancelled[0] = True
            try:
                # process tasks per dataset
                self.signal_begin.emit()
                for dataset in datasets:
                    task_queue = [task for task in task_queue_all if task.dataset == dataset]
                    expressions = list(set(expression for task in task_queue for expression in task.expressions_all))

                    for task in task_queue:
                        task._results = []
                        task.signal_progress.emit(0)
                    self.server.execute_queue(task_queue)
                    self._is_executing = False
            except:
                # on any error we flush the task queue
                self.signal_cancel.emit()
                logger.exception("error in task, flush task queue")
                raise
            logger.debug("executing took %r seconds" % (time.time() - t0))
            # while processing the self.task_queue, new elements will be added to it, so copy it
            logger.debug("cancelled: %r", cancelled)
            if cancelled[0]:
                logger.debug("execution aborted")
                task_queue = task_queue_all
                for task in task_queue:
                    # task._result = task.reduce(task._results)
                    # task.reject(UserAbort("cancelled"))
                    # remove references
                    task._result = None
                    task._results = None
            else:
                task_queue = task_queue_all
                for task in task_queue:
                    logger.debug("fulfill task: %r", task)
                    if not task.cancelled:
                        task._result = task.reduce(task._results)
                        task.fulfill(task._result)
                        # remove references
                    task._result = None
                    task._results = None
                self.signal_end.emit()
                # if new tasks were added as a result of this, execute them immediately
                # TODO: we may want to include infinite recursion protection
                self._is_executing = False
                if len(self.task_queue) > 0:
                    logger.debug("task queue not empty.. start over!")
                    self.execute()
        finally:
            self._is_executing = False


if __name__ == "__main__":
    import vaex
    import sys
    vaex.set_log_level_debug()
    server = vaex.server(sys.argv[1], port=int(sys.argv[2]))
    datasets = server.datasets()
    print(datasets)
    dataset = datasets[0]
    dataset = vaex.example()
    print(dataset("x").minmax())
    dataset.select("x < 0")
    print(dataset.selected_length(), len(dataset))
    print(dataset("x").selected().is_masked)
    print(dataset("x").selected().minmax())



# -*- coding: utf-8 -*-
import logging as logging
import numpy as np
import math
import functools
import vaex.vaexfast
import vaex.utils
from vaex.utils import Timer

logger = logging.getLogger("vaex.kld")


def mutual_information(data):
    Q = vaex.utils.disjoined(data)
    P = data

    P = P / P.sum()
    Q = Q / Q.sum()
    mask = (P > 0) & (Q > 0)
    information = np.sum(P[mask] * np.log(P[mask] / Q[mask]))  # * np.sum(dx)
    return information


def kl_divergence(P, Q, axis=None):
    P = P / P.sum(axis=axis)
    Q = Q / Q.sum(axis=axis)
    # mask = (P > 0) & (Q>0)
    # information = np.sum(P[mask] * np.log(P[mask]/Q[mask]), axis=axis)# * np.sum(dx)
    # mask = (P > 0) & (Q>0)
    information = np.sum(P * np.log(P / Q), axis=axis)  # * np.sum(dx)
    return information


class KlDivergenceShuffle(object):
    def __init__(self, dataset, pairs, gridsize=128):
        self.dataset = dataset
        self.pairs = pairs
        self.dimension = len(self.pairs[0])
        self.logger = logger.getLogger('kld')
        self.gridsize = gridsize
        logger.debug("dimension: %d, pairs: %s" % (self.dimension, self.pairs))

    def get_jobs(self):
        def job(pair):
            pass


def to_disjoined(counts):
    shape = counts.shape
    assert len(counts.shape) == 2
    counts_0 = counts.sum(axis=1).reshape((shape[0], 1))
    counts_1 = counts.sum(axis=0).reshape((1, shape[1]))
    counts_disjoined = counts_0 * counts_1
    return counts_disjoined


def kld_shuffled(columns, Ngrid=128, datamins=None, datamaxes=None, offset=1):
    if datamins is None:
        datamins = np.array([np.nanmin(column) for column in columns])
    if datamaxes is None:
        datamaxes = np.array([np.nanmax(column) for column in columns])
    dim = len(columns)
    counts = np.zeros((Ngrid, ) * dim, dtype=np.float64)
    counts_shuffled = np.zeros((Ngrid, ) * dim, dtype=np.float64)
    D_kl = -1
    if len(columns) == 2:
        x, y = columns
        # print x
        # print y
        print((x, y, counts, counts_shuffled, datamins[0], datamaxes[0], datamins[1], datamaxes[1], offset))
        try:
            vaex.histogram.hist2d_and_shuffled(x, y, counts, counts_shuffled, datamins[0], datamaxes[0], datamins[1], datamaxes[1], offset)
        except:
            args = [x, y, counts, counts_shuffled, datamins[0], datamaxes[0], datamins[1], datamaxes[1], offset]
            sig = [numba.dispatcher.typeof_pyval(a) for a in args]
            print(sig)
            raise

        print(("counts", sum(counts)))
        deltax = [float(datamaxes[i] - datamins[i]) for i in range(dim)]
        dx = np.array([deltax[d] / counts.shape[d] for d in range(dim)])
        density = counts / np.sum(counts)  # * np.sum(dx)
        density_shuffled = counts_shuffled / np.sum(counts_shuffled)  # * np.sum(dx)
        mask = (density_shuffled > 0) & (density > 0)
        # print density
        D_kl = np.sum(density[mask] * np.log(density[mask] / density_shuffled[mask]))  # * np.sum(dx)
        # if D_kl < 0:
        # import pdb
        # pdb.set_trace()
    return D_kl


def kld_shuffled_grouped(dataset, range_map, pairs, feedback=None, size_grid=32, use_mask=True, bytes_max=int(1024**3 / 2)):
    dimension = len(pairs[0])
    bytes_per_grid = size_grid ** dimension * 8  # 8==sizeof (double)
    grids_per_iteration = min(128, int(bytes_max / bytes_per_grid))
    iterations = int(math.ceil(len(pairs) * 1. / grids_per_iteration))
    jobsManager = vaex.dataset.JobsManager()
    ranges = [None] * len(pairs)
    D_kls = []

    class Wrapper(object):
        pass
    wrapper = Wrapper()
    wrapper.N_done = 0

    counts = np.zeros((grids_per_iteration,) + (size_grid,) * dimension, dtype=np.float64)
    # counts_shuffled = np.zeros((grids_per_iteration,) + (size_grid,) * dimension, dtype=np.float64)

    N_total = len(pairs) * len(dataset) * dimension + len(pairs) * counts.size

    logger.debug("{iterations} iterations with {grids_per_iteration} grids per iteration".format(**locals()))
    for part in range(iterations):
        if part > 0:  # next iterations reset the counts
            counts.reshape(-1)[:] = 0
            # counts_shuffled.reshape(-1)[:] = 0
        i1, i2 = part * grids_per_iteration, (part + 1) * grids_per_iteration
        if i2 > len(pairs):
            i2 = len(pairs)
        n_grids = i2 - i1
        logger.debug("part {part} of {iterations}, from {i1} to {i2}".format(**locals()))

        def grid(info, *blocks, **kwargs):
            index = kwargs["index"]
            if use_mask and dataset.mask is not None:
                mask = dataset.mask[info.i1:info.i2]
                blocks = [block[mask] for block in blocks]
            else:
                mask = None
            ranges = []
            minima = []
            maxima = []
            for dim in range(dimension):
                ranges += list(range_map[pairs[index][dim]])
                mi, ma = range_map[pairs[index][dim]]
                minima.append(mi)
                maxima.append(ma)
            if len(blocks) == 2:
                print(("mask", mask))
                vaex.vaexfast.histogram2d(blocks[0], blocks[1], None, counts[index], *(ranges + [0, 0]))
                # vaex.vaexfast.histogram2d(blocks[0], blocks[1], None, counts_shuffled[index], *(ranges + [1, 0]))
            if len(blocks) == 3:
                vaex.vaexfast.histogram3d(blocks[0], blocks[1], blocks[2], None, counts[index], *(ranges + [0, 0, 0]))
                # vaex.vaexfast.histogramNd(list(blocks), None, counts[index], minima, maxima)
                # for i in range(5):
                # vaex.vaexfast.histogram3d(blocks[0], blocks[1], blocks[2], None, counts_shuffled[index], *(ranges + [2+i,1+i,0]))
            if len(blocks) > 3:
                vaex.vaexfast.histogramNd(list(blocks), None, counts[index], minima, maxima)
                # for i in range(5):
                # vaex.vaexfast.histogram3d(blocks[0], blocks[1], blocks[2], None, counts_shuffled[index], *(ranges + [2+i,1+i,0]))
            if feedback:
                wrapper.N_done += len(blocks[0]) * dimension
                if feedback:
                    cancel = feedback(wrapper.N_done * 100. / N_total)
                    if cancel:
                        raise Exception("cancelled")

        for index, pair in zip(list(range(i1, i2)), pairs[i1:i2]):
            logger.debug("add job %r %r" % (index, pair))
            jobsManager.addJob(0, functools.partial(grid, index=index - i1), dataset, *pair)
        jobsManager.execute()

        with Timer("D_kl"):
            for i in range(n_grids):
                if 0:
                    deltax = [float(range_map[pairs[i][d]][1] - range_map[pairs[i][d]][0]) for d in range(dimension)]
                    dx = np.array([deltax[d] / counts[i].shape[d] for d in range(dimension)])
                    density = counts[i] / np.sum(counts[i])  # * np.sum(dx)
                    counts_shuffled = to_disjoined(counts[i])
                    density_shuffled = counts_shuffled / np.sum(counts_shuffled)  # * np.sum(dx)
                    mask = (density_shuffled > 0) & (density > 0)
                    print(("mask sum", np.sum(mask)))
                    print(("mask sum", np.sum((counts_shuffled > 0) & (counts[i] > 0))))
                    # print density
                    D_kl = np.sum(density[mask] * np.log(density[mask] / density_shuffled[mask]))  # * np.sum(dx)
                else:
                    D_kl = mutual_information(counts[i])
                D_kls.append(D_kl)
                if feedback:
                    wrapper.N_done += counts.size
                    cancel = feedback(wrapper.N_done * 100. / N_total)
                    if cancel:
                        return None
                        # raise Exception, "cancelled"
    return D_kls


if __name__ == "__main__":
    import vaex.dataset
    import vaex.files
    from optparse import OptionParser
    parser = OptionParser()  # usage="")

    # parser.add_option("-n", "--name",
    #                 help="dataset name [default=%default]", default="data", type=str)
    # parser.add_option("-o", "--output",
    #                 help="dataset output filename [by default the suffix of input filename will be replaced by hdf5]", default=None, type=str)
    (options, args) = parser.parse_args()
    if len(args) > 0:
        filename = args[0]
    else:
        filename = "gaussian4d-1e7.hdf5"

    path = vaex.files.get_datafile(filename)
    dataset = vaex.dataset.Hdf5MemoryMapped(path)
    # for column_name in dataset.column_names:
    # print column_name

    subspace = dataset.subspace(1)[:3] * dataset.subspace(1)[:3]
    # subspace = Subspace(dataset, [("x",), ("y",), ("z",)]



import logging
import os

import six
import numpy as np

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import vaex.strings

logger = logging.getLogger("vaex.column")


class Column(object):
    def tolist(self):
        return self.to_numpy().tolist()


class ColumnVirtualRange(Column):
    def __init__(self, start, stop, step=1, dtype=None):
        self.start = start
        self.stop = stop
        self.step = step
        self.dtype = np.dtype(dtype)
        self.shape = (self.__len__(),)

    def __len__(self):
        return (self.stop - self.start) // self.step

    def __getitem__(self,  slice):
        start, stop, step = slice.start, slice.stop, slice.step
        return np.arange(self.start + start, self.start + stop, step, dtype=self.dtype)

    def trim(self, i1, i2):
        return ColumnVirtualRange(self.start + i1 * self.step, self.start + i2 * self.step, self.step, self.dtype)


class ColumnSparse(object):
    def __init__(self, matrix, column_index):
        self.matrix = matrix
        self.column_index = column_index
        self.shape = self.matrix.shape[:1]
        self.dtype = self.matrix.dtype

    def __len__(self):
        return self.shape[0]

    def __getitem__(self, slice):
        # not sure if this is the fastest
        return self.matrix[slice, self.column_index].A[:,0]

class ColumnNumpyLike(Column):
    """Wraps a numpy like object (like hdf5 dataset) into a column vaex understands"""
    def __init__(self, ar):
        self.ar = ar  # this should behave like a numpy array

    def __len__(self):
        return len(self.ar)

    def trim(self, i1, i2):
        return type(self)(self.ar[i1:i2])

    def __getitem__(self, slice):
        return self.ar[slice]

    def __setitem__(self, slice, value):
        self.ar[slice] = value

class ColumnIndexed(Column):
    def __init__(self, df, indices, name):
        self.df = df
        self.indices = indices
        self.name = name
        self.dtype = self.df.dtype(name)
        self.shape = (len(indices),)
        max_index = self.indices.max()
        assert max_index < self.df._length_unfiltered

    def __len__(self):
        return len(self.indices)

    def trim(self, i1, i2):
        return ColumnIndexed(self.df, self.indices[i1:i2], self.name)

    def __getitem__(self, slice):
        start, stop, step = slice.start, slice.stop, slice.step
        start = start or 0
        stop = stop or len(self)
        assert step in [None, 1]
        indices = self.indices[start:stop]
        ar_unfiltered = self.df.columns[self.name]
        ar = ar_unfiltered[indices]
        if np.ma.isMaskedArray(indices):
            mask = self.indices.mask[start:stop]
            return np.ma.array(ar, mask=mask)
        else:
            return ar


class ColumnConcatenatedLazy(Column):
    def __init__(self, dfs, column_name):
        self.dfs = dfs
        self.column_name = column_name
        dtypes = [df.dtype(column_name) for df in dfs]
        self.is_masked = any([df.is_masked(column_name) for df in dfs])
        if self.is_masked:
            self.fill_value = dfs[0].columns[self.column_name].fill_value
        # np.datetime64 and find_common_type don't mix very well
        any_strings = any([dtype == str_type for dtype in dtypes])
        if any_strings:
            self.dtype = str_type
        else:
            if all([dtype.type == np.datetime64 for dtype in dtypes]):
                self.dtype = dtypes[0]
            else:
                if all([dtype == dtypes[0] for dtype in dtypes]):  # find common types doesn't always behave well
                    self.dtype = dtypes[0]
                if  any([dtype.kind in 'SU' for dtype in dtypes]):  # strings are also done manually
                    if all([dtype.kind in 'SU' for dtype in dtypes]):
                        index = np.argmax([dtype.itemsize for dtype in dtypes])
                        self.dtype = dtypes[index]
                    else:
                        index = np.argmax([df.columns[self.column_name].astype('O').astype('U').dtype.itemsize for df in dfs])
                        self.dtype = dfs[index].columns[self.column_name].astype('O').astype('U').dtype
                else:
                    self.dtype = np.find_common_type(dtypes, [])
                logger.debug("common type for %r is %r", dtypes, self.dtype)
        self.shape = (len(self), ) + self.dfs[0].evaluate(self.column_name, i1=0, i2=1).shape[1:]
        for i in range(1, len(dfs)):
            shape_i = (len(self), ) + self.dfs[i].evaluate(self.column_name, i1=0, i2=1).shape[1:]
            if self.shape != shape_i:
                raise ValueError("shape of of column %s, array index 0, is %r and is incompatible with the shape of the same column of array index %d, %r" % (self.column_name, self.shape, i, shape_i))

    def __len__(self):
        return sum(len(ds) for ds in self.dfs)

    def __getitem__(self, slice):
        start, stop, step = slice.start, slice.stop, slice.step
        start = start or 0
        stop = stop or len(self)
        assert step in [None, 1]
        dtype = self.dtype
        if dtype == str_type:
            dtype = 'O'  # we store the strings in a dtype=object array
        dfs = iter(self.dfs)
        current_df = next(dfs)
        offset = 0
        # print "#@!", start, stop, [len(df) for df in self.dfs]
        while start >= offset + len(current_df):
            # print offset
            offset += len(current_df)
            # try:
            current_df = next(dfs)
            # except StopIteration:
            # logger.exception("requested start:stop %d:%d when max was %d, offset=%d" % (start, stop, offset+len(current_df), offset))
            # raise
            #   break
        # this is the fast path, no copy needed
        if stop <= offset + len(current_df):
            if current_df.filtered:  # TODO this may get slow! we're evaluating everything
                warnings.warn("might be slow, you have concatenated dfs with a filter set")
            return current_df.evaluate(self.column_name, i1=start - offset, i2=stop - offset)
        else:
            if self.is_masked:
                copy = np.ma.empty(stop - start, dtype=dtype)
                copy.fill_value = self.fill_value
            else:
                copy = np.zeros(stop - start, dtype=dtype)
            copy_offset = 0
            # print("!!>", start, stop, offset, len(current_df), current_df.columns[self.column_name])
            while offset < stop:  # > offset + len(current_df):
                # print(offset, stop)
                if current_df.filtered:  # TODO this may get slow! we're evaluating everything
                    warnings.warn("might be slow, you have concatenated DataFrames with a filter set")
                part = current_df.evaluate(self.column_name, i1=start-offset, i2=min(len(current_df), stop - offset))
                # print "part", part, copy_offset,copy_offset+len(part)
                copy[copy_offset:copy_offset + len(part)] = part
                # print copy[copy_offset:copy_offset+len(part)]
                offset += len(current_df)
                copy_offset += len(part)
                start = offset
                if offset < stop:
                    current_df = next(dfs)
            return copy

try:
    str_type = unicode
except:
    str_type = str

use_c_api = True

class ColumnString(Column):
    pass

class ColumnStringArray(Column):
    """Wraps a numpy array with dtype=object, containing all strings"""
    def __init__(self, array):
        # assert vaex.strings.all_strings(array, allow)
        self.array = array
        self.string_sequence = vaex.strings.StringArray(self.array)

    def __getitem__(self, slice):
        return ColumnString(self.array[slice])

    def to_numpy(self):
        return self.array

def _to_string(string_bytes):
    if six.PY2:
        return unicode(buffer(string_bytes))
    else:
        return bytes(memoryview(string_bytes)).decode('utf8')

def _to_string_sequence(x):
    if isinstance(x, ColumnString):
        return x.string_sequence
    elif isinstance(x, np.ndarray):
        if x.dtype == 'O':
            return vaex.strings.StringArray(x)
        elif x.dtype.kind in 'US':
            x = x.astype('O')
            return vaex.strings.StringArray(x)
        else:
            raise ValueError('unsupported dtype ' +str(x.dtype))
    else:
        raise ValueError('not a ColumnString or ndarray: ' + str(x))

def _to_string_list_sequence(x):
    if isinstance(x, vaex.strings.StringListList):
        return x
    else:
        raise ValueError('not a StringListList')

class ColumnStringArrow(ColumnString):
    """Column that unpacks the arrow string column on the fly"""
    def __init__(self, indices, bytes, length=None, offset=0, string_sequence=None, null_bitmap=None):
        self.indices = indices
        self.offset = offset  # to avoid memory copies in trim
        self.bytes = bytes
        self.dtype = str_type
        self.length = length if length is not None else len(indices) - 1
        self.shape = (self.__len__(),)
        self.nbytes = self.bytes.nbytes + self.indices.nbytes
        self.null_bitmap = null_bitmap
        if string_sequence is None:
            if self.indices.dtype.kind == 'i' and self.indices.dtype.itemsize == 8:
                string_type = vaex.strings.StringList64
            elif self.indices.dtype.kind == 'i' and self.indices.dtype.itemsize == 4:
                string_type = vaex.strings.StringList32
            else:
                raise ValueError('unsupported index type' + str(self.indices.dtype))
            if self.null_bitmap is not None:
                self.string_sequence = string_type(self.bytes, self.indices, self.length, self.offset, null_bitmap)
            else:
                self.string_sequence = string_type(self.bytes, self.indices, self.length, self.offset)
        else:
            self.string_sequence = string_sequence

    def __len__(self):
        return self.length

    def __getitem__(self, slice):
        if isinstance(slice, int):
            return self.string_sequence.get(slice)
        elif isinstance(slice, np.ndarray):
            return type(self).from_string_sequence(self.string_sequence.index(slice))
        else:
            start, stop, step = slice.start, slice.stop, slice.step
            start = start or 0
            stop = stop or len(self)
            assert step in [None, 1]
            return self.trim(start, stop)

    def to_numpy(self):
        return self.string_sequence.to_numpy()

    def trim(self, i1, i2):
        return type(self).from_string_sequence(self.string_sequence.slice(i1, i2))

    @classmethod
    def from_string_sequence(cls, string_sequence):
        s = string_sequence
        return cls(s.bytes, s.indices, s.length, s.offset, string_sequence=s, null_bitmap=s.null_bitmap)

    def _zeros_like(self):
        return ColumnStringArrow(np.zeros_like(self.indices), np.zeros_like(self.bytes), self.length, null_bitmap=self.null_bitmap)

    def get_mask(self):
        return self.string_sequence.mask()
